/*******************************************************************************

  Copyright(c) 2003 - 2005 Promise Technology, Inc. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information:
  Promise Technology, Inc.
  <support@promise.com.tw>	[TAIWAN]
  <support@promise.com>		[U.S.A]
  <support-china@promise.com>	[CHINA]

*******************************************************************************/

#include <stdarg.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/byteorder.h>
#include <asm/uaccess.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/blkdev.h>
#include "scsi.h"
#include <scsi/scsi_host.h>
#include <scsi/scsi.h>
#include <linux/reboot.h>
#else	/* 2.6.0 */
#include <linux/blk.h>
#include <linux/tqueue.h>
#include <linux/wrapper.h>
#include "scsi/sd.h"
#include "scsi/scsi.h"
#include "scsi/hosts.h"
#endif	/* 2.6.0 */
#include <linux/types.h>
#include <linux/hdreg.h>
#include <linux/interrupt.h>

#include <linux/fs.h>
#include <linux/timer.h>
#include <scsi/sg.h>
#include <linux/module.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,1)
#include <drivers/scsi/sd.h>
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <drivers/scsi/scsi.h>
#include <drivers/scsi/hosts.h>
#endif
#include "cam/cam_con.h"
#include "cam/cam_def.h"
#include "cam/cam_exp.h"

#include "pdc-ulsata2.h"
#define	ulsata2_present	0x1
#define	ulsata2_IN_INTR	0x2
/*
 * function prototype
 */
static void hexdump(u8 *x, int len);
static int ulsata2_ata_rw(Scsi_Cmnd *SCpnt);
static void ulsata2_init_device(unsigned long drv);
static int ulsata2_halt(struct notifier_block *nb, unsigned long event, void *buf);
static int ulsata2dev_open(struct inode *inodep, struct file *filep);
static int ulsata2dev_release(struct inode *inodep, struct file *filep);
static int ulsata2dev_ioctl(struct inode *inodep, struct file *filep, unsigned int req, unsigned long arg);

void SwapBox_CtrlLED( U8 channel, U8 Status, U8 Flash, U8 Type );
void SwapBox_CtrlLEDEmptyDevices( U8 AdapterId, U8 Status, U8 Flash, U8 Type );
struct timer_list	swapbox_timer;
int	swapbox_exist = 0 ;
static unsigned short p_LED[MAXCHANNEL];


#ifdef	_DEBUG_
#define	DBG(str,args...)	printk( "%s : "str,__FUNCTION__,##args )
#else
#define	DBG(str,args...)
#endif

/* Do not export any gobal variable in system symbol table */
//EXPORT_NO_SYMBOLS;



#define minmax(a,b)	((a)<(b) ? (a):(b))

/* linux kernel version related */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#define KERNEL_VERSION_24x
#else
static int pci_enable_device(struct pci_dev *dev) {}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,3)
static int pci_set_dma_mask(struct pci_dev *dev, u64 mask) {}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,47)
#define SCSI_DYNAMIC_DMA_MAPPING
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,18)
typedef u32 dma_addr_t;		/* 2.2.18 defines dma_addr_t */
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,1)
#define KERNEL_VERSION_25x
#endif

/*
 * global data structure
 */
/* /proc directory entry for kernel v2.2 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,27)
static struct proc_dir_entry proc_scsi_ulsata2 = {
    PROC_SCSI_NOT_PRESENT, 9, PDCNAME,
    S_IFDIR | S_IRUGO | S_IXUGO, 2
};
#endif

/*
 * The file operations structure for the ioctl interface of the driver
 */
static struct file_operations ulsata2_fops = {
#ifndef KERNEL_VERSION_25x
	open: 		ulsata2dev_open,
	release:	ulsata2dev_release,
	ioctl:		ulsata2dev_ioctl,
#else
	.owner	=	THIS_MODULE,
	.ioctl	=	ulsata2dev_ioctl,
	.open	=	ulsata2dev_open,
	.release =	ulsata2dev_release
#endif
};


/* notifier block to get a notifier on system shutdown/halt/poweroff */
static struct notifier_block ulsata2_notifier = {
	ulsata2_halt, NULL, 0
};

ulsata2_adapter_t	ulsata2_adapter[MAXADAPTER];
ulsata2_channel_t	ulsata2_channel[MAXCHANNEL];
ulsata2_drive_t		ulsata2_drive[MAXDRIVE];
ulsata2_timer_t		ulsata2_timer[MAXTIMER];
//ulsata2_req_t		*ulsata2_req[QLENGTH+1], *PoolH, *PoolF;
ulsata2_req_t		*ulsata2_req[(QLENGTH*2)-1], *PoolH, *PoolF; //ulsata2_req[127]
spinlock_t			CAM_spinlock;
MODULE_TIMEOUT 		*CamTimeoutTable;
	
	

/* user configurable values */
int	ut_adapters=0;
int	ATAPI_BUF_Check = 0;		//for Suse9.2
//unsigned char errTimePCH[MAXCHANNEL];
static int atapi = 1;			/* en/disable atapi support */
static int debug = 0;			/* The debug level */
static DECLARE_WAIT_QUEUE_HEAD(wait);
static int ulsata2_unlock;
static int ulsata2_num_adapters;		/* total adapters */
static int ulsata2_major;		/* char dev major number */

#ifndef SCSI_DATA_WRITE
#define SCSI_DATA_WRITE 	1
#endif
#ifndef SCSI_RESET_SUCCESS
#define SCSI_RESET_SUCCESS 	2
#endif
#ifndef SCSI_RESET_ERROR
#define SCSI_RESET_ERROR 	6
#endif
#ifndef SCSI_ABORT_NOT_RUNNING
#define SCSI_ABORT_NOT_RUNNING	5
#endif
#ifndef SCSI_ABORT_SUCCESS
#define SCSI_ABORT_SUCCESS	1
#endif
#ifndef scsi_to_pci_dma_dir
#define scsi_to_pci_dma_dir(scsi_dir) ((int)(scsi_dir)) 
#endif
//#ifndef SCSI_DATA_READ
//#define SCSI_DATA_READ 		2
//#endif


/*
 * Deal with DMA mapping/unmapping.
 */
#ifndef SCSI_DYNAMIC_DMA_MAPPING

#define __unmap_scsi_data(pdev, cmd)	  do {; } while (0)
#define __map_scsi_single_data(pdev, cmd) (virt_to_bus((cmd)->request_buffer))
#define __map_scsi_sg_data(pdev, cmd)	  ((cmd)->use_sg)
#if 0
#define __sync_scsi_data(pdev, cmd)	  do {; } while (0)
#endif

#define scsi_sg_dma_address(sc)		  virt_to_bus((sc)->address)
#define scsi_sg_dma_len(sc)		  ((sc)->length)

#else

/* To keep track of the dma mapping (sg/single) that has been set */
#define __data_mapped(cmd)	(cmd)->SCp.phase
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,13)
#define __data_mapping(cmd)	(cmd)->SCp.have_data_in
#else
#define __data_mapping(cmd)	(cmd)->SCp.dma_handle
#endif

static void __unmap_scsi_data(struct pci_dev *pdev, Scsi_Cmnd *SCpnt)
{
	int dma_dir = scsi_to_pci_dma_dir(SCpnt->sc_data_direction);

	switch(__data_mapped(SCpnt)) {
	case 1:
#ifdef BLK_BOUNCE_HIGH /* Hank */
		pci_unmap_page(pdev, __data_mapping(SCpnt), (size_t)SCpnt->request_bufflen, dma_dir);
#else
		pci_unmap_single(pdev, __data_mapping(SCpnt), (size_t)SCpnt->request_bufflen, dma_dir);
#endif
		break;
	case 2:
		pci_unmap_sg(pdev, SCpnt->buffer, (int)SCpnt->use_sg, dma_dir);
		break;
	}
	__data_mapped(SCpnt) = 0;
}

static dma_addr_t __map_scsi_single_data(struct pci_dev *pdev, Scsi_Cmnd *SCpnt)
{
	dma_addr_t mapping;
	int dma_dir = scsi_to_pci_dma_dir(SCpnt->sc_data_direction);

	if (SCpnt->request_bufflen == 0)
		return 0;

#ifdef BLK_BOUNCE_HIGH 	/* Hank */
	mapping = pci_map_page(pdev, virt_to_page(SCpnt->request_buffer), 
			((unsigned long)SCpnt->request_buffer & ~PAGE_MASK), 
			(size_t)SCpnt->request_bufflen, dma_dir); 
#else
	mapping = pci_map_single(pdev, SCpnt->request_buffer, (size_t)SCpnt->request_bufflen, dma_dir);
	//pci_map_single(struct pci_dev *hwdev, void *ptr, size_t size, int direction)
#endif

	__data_mapped(SCpnt) = 1;
	__data_mapping(SCpnt) = mapping;

	return mapping;
}

static int __map_scsi_sg_data(struct pci_dev *pdev, Scsi_Cmnd *SCpnt)
{
	int use_sg;
	int dma_dir = scsi_to_pci_dma_dir(SCpnt->sc_data_direction);

	if (SCpnt->use_sg == 0)
		return 0;

	use_sg = pci_map_sg(pdev, SCpnt->buffer, SCpnt->use_sg, dma_dir);
	__data_mapped(SCpnt) = 2;
	
	#ifdef _64BIT_
	__data_mapping(SCpnt) = (dma_addr_t)use_sg;
	#else
	__data_mapping(SCpnt) = use_sg;
	#endif

	return use_sg;
}

#define scsi_sg_dma_address(sc)		sg_dma_address((sc))
#define scsi_sg_dma_len(sc)		sg_dma_len((sc))

#endif	/* SCSI_DYNAMIC_DMA_MAPPING */

#define unmap_scsi_data(np, cmd)	__unmap_scsi_data((np)->pci_dev, (cmd))
#define map_scsi_single_data(np, cmd)	__map_scsi_single_data((np)->pci_dev, (cmd))
#define map_scsi_sg_data(np, cmd)	__map_scsi_sg_data((np)->pci_dev, (cmd))

/*
 * dump hex data
 */
static void hexdump(u8 *x, int len)
{
	int i;

	printk("[ ");
	for (i = 0; i < len; i++)
		printk("%x ", x[i]);
	printk("]\n");
}

int init_pci_resource(ulsata2_adapter_t *pada ,
					struct pci_dev* pdev)
{
	int i;
	//CAM_INFO *pcam_info = pada->cam_info;
	ADAPTER_CONFIG_INFO *pcam_ada = pada->cam_ada;
	
	//CAM_GetInfo(pcam_info);
	printk("Start to init_pci_resource\n");
	for(i=0; i<6; i++){
		if (pci_resource_len(pdev, i) == 0){
			pada->base[i] = 0;
			continue;
		}			
		if(pci_resource_flags(pdev, i) & IORESOURCE_IO) 
		{	pada->base[i] = (unsigned int)pci_resource_start(pdev, i);
			pada->range[i] = (unsigned int)pci_resource_len(pdev, i);
			printk("BA[%d]=%X--IO\n",i,pada->base[i]);
			pcam_ada->BaseAddress[i] = (unsigned long)pada->base[i];
		}
		if(pci_resource_flags(pdev, i) & IORESOURCE_MEM)
		{	pcam_ada->BaseAddress[i]  = (unsigned long)ioremap((unsigned long)pci_resource_start(pdev,i), 
										(unsigned long)pci_resource_len(pdev,i) );
			pada->base[i]= (unsigned int)pcam_ada->BaseAddress[i];						
			pada->range[i] = (unsigned int)pci_resource_len(pdev, i);	
			
			printk("BA[%d]=%X--MMIO\n",i,pada->base[i]);							
		}
		

	}
	
	printk("Exit to init_pci_resource\n");
	return 0;

}
/*
 * initial global data structure
 */
int ulsata2_init_gds(void)
{
	unsigned char i,j;
	if((ulsata2_adapter[0].present)&ulsata2_present)
		return 0;
	else
		printk("Init_GDS_Here!!!!\n");		
	
	spin_lock_init(&CAM_spinlock);	
	CamTimeoutTable = kmalloc(sizeof(MODULE_TIMEOUT), GFP_ATOMIC);
	/* initialize ulsata2_adapter[] */
	for (i = 0; i < MAXADAPTER; i++) {
		ulsata2_adapter_t *pada = &ulsata2_adapter[i];
		memset(pada,0,sizeof(ulsata2_adapter_t));
		pada->id = i;
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)	
		pada->ulsata2_lock=&CAM_spinlock;
	#endif		
		for (j = 0; j < ChPAda; j++) {
			pada->pchannel[j] = &ulsata2_channel[i*ChPAda+j];
		}
		pada->cam_ada = kmalloc(sizeof(ADAPTER_CONFIG_INFO), GFP_ATOMIC);
		memset(pada->cam_ada, 0, sizeof(ADAPTER_CONFIG_INFO));
		pada->cam_info = kmalloc(sizeof(CAM_INFO), GFP_ATOMIC);
		memset(pada->cam_info, 0, sizeof(CAM_INFO));
	}


	/* initialize ulsata2_channel[] */
	for (i = 0; i < MAXCHANNEL; i++) {
		ulsata2_channel_t *pchn = &ulsata2_channel[i];
		memset(pchn,0,sizeof(ulsata2_channel_t));
		pchn->id = i;
		pchn->padapter = &ulsata2_adapter[i/ChPAda];
		for (j = 0 ; j < DrvPCh; j++) {
			pchn->pdrive[j] = &ulsata2_drive[i*DrvPCh+j];
		}
		pchn->cam_chn = kmalloc(sizeof(CHNL_CONFIG), GFP_ATOMIC);
		memset(pchn->cam_chn,0 ,sizeof(CHNL_CONFIG));
		//errTimePCH[i]=0;
	}
	/* initialize ulsata2_drive[] */
	for (i = 0; i < MAXDRIVE; i++) {
		ulsata2_drive_t *pdrv = &ulsata2_drive[i];
		memset(pdrv,0,sizeof(ulsata2_drive_t));
		ulsata2_drive[i].pchannel = &ulsata2_channel[i/DrvPCh];
		pdrv->cam_dev = kmalloc(sizeof(DEV_CONFIG), GFP_ATOMIC);
		memset(pdrv->cam_dev, 0, sizeof(DEV_CONFIG));
	}

	/* initialize ulsata2_timer[] */
	for ( i = 0; i < MAXTIMER; i++) {
		ulsata2_timer[i].status = camFREE;
	}

	/* initialize submit ata request queue */
	//for ( i = 0 ; i < (QLENGTH +1); i++ ) {
	for ( i = 0 ; i < (QLENGTH*2)-1; i++ ) {	
		ulsata2_req[i] = kmalloc(sizeof(ulsata2_req_t), GFP_ATOMIC|__GFP_DMA);
		if (!ulsata2_req[i]) {
		       	printk(KERN_ERR "%s:[error] can't allocate enough memory for queue\n",PDCNAME);
			//ulsata2_cleanup(0);
			//ulsata2_cleanup(1);
			//ulsata2_cleanup2();
		       	return -1;
		}
		memset(ulsata2_req[i],0, sizeof(ulsata2_req_t));
		if ( i == 0 )
			ulsata2_req[i]->next = NULL;
		else
			ulsata2_req[i]->next = ulsata2_req[i-1];
	}
	PoolH = ulsata2_req[(QLENGTH-1)*2];
	PoolF = ulsata2_req[0];

	return(0);
}

/*
 * interrupt handler
 */
#ifndef KERNEL_VERSION_25x
static void ulsata2_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#else
int ulsata2_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
	unsigned long flags	= 0;
	unsigned long irqhandle = 0;
	ulsata2_adapter_t *pada	= (ulsata2_adapter_t *) dev_id;
	unsigned char ua;
	//spinlock_t *host_lock;	//Ethan
	int handled = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	int irqret;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,3)
	host_lock = &io_request_lock;
#else
	//host_lock = pada->host->host_lock; //Ethan
#endif

	/* check adapter data */
	if (!pada){
#ifndef KERNEL_VERSION_25x
		return;
#else
		return IRQ_NONE;
#endif
	}

	/* set adapter bit according to irq */
	for (ua = 0; ua < MAXADAPTER; ua++) {
		if((ulsata2_adapter[ua].present)&ulsata2_present) {
			if (ulsata2_adapter[ua].irq == irq)
				irqhandle |= 1<<ua;

		}
	}

	/* not our irq */
	if (!irqhandle) {
#ifndef KERNEL_VERSION_25x
		return;
#else
		return IRQ_NONE;
#endif
	}

	/* debug */
	if (debug & LOGATA) {
		printk(KERN_DEBUG "%s:[D] ISR handle=%#lx\n", 
				PDCNAME, irqhandle);
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)	
	spin_lock_irqsave(pada->ulsata2_lock, flags);
#else	
	spin_lock_irqsave(&CAM_spinlock, flags);
#endif	
	handled = 1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	irqret = 
#endif	
	CAM_ISR(irqhandle);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)		
	spin_unlock_irqrestore(pada->ulsata2_lock, flags);
#else	
	spin_unlock_irqrestore(&CAM_spinlock, flags);
#endif	
#ifndef KERNEL_VERSION_25x
	return;
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)	//Ethan
	return(irqret);
#else
	return IRQ_RETVAL(handled);
#endif	
#endif
}

/*
 * initialize ioport and irq
 */
int ulsata2_init_ioport(unsigned char ada)
{
	unsigned char	uc;
	ulsata2_adapter_t *pada = &ulsata2_adapter[ada];
	struct pci_dev	*ulsata2_pci_dev = pada->pci_dev;
	unsigned short	command;
	CAM_INFO *pcam_info = pada->cam_info;
	ADAPTER_CONFIG_INFO *pcam_ada = pada->cam_ada;
	
	CAM_GetInfo(pcam_info);
	/* read base address and get ioport range */
	for (uc = 0; uc < 6; uc++) {
	       	unsigned int	base = 0 , range = 0;
		pci_read_config_dword(ulsata2_pci_dev, PCI_BASE_ADDRESS_0+4*uc, &base);
		pci_write_config_dword(ulsata2_pci_dev,PCI_BASE_ADDRESS_0+4*uc,0xffffffff);
		pci_read_config_dword(ulsata2_pci_dev, PCI_BASE_ADDRESS_0+4*uc, &range);
		range = ~range + 1;
		pada->range[uc] = range + ((uc < 3) ? 1 : 0);

		pci_write_config_dword(ulsata2_pci_dev,PCI_BASE_ADDRESS_0+4*uc,base);
		pada->base[uc] = base & PCI_BASE_ADDRESS_IO_MASK;
		/* DBG("base[%d]=%#x\n",uc,pada->base[uc]); */

		/* assign BASE address for CAM */
#ifdef _MMIO_
		if (pcam_info->VirMemBase & bit(uc) ) {
			pcam_ada->BaseAddress[uc] = (unsigned long)ioremap((unsigned long)pada->base[uc], (unsigned long)pada->range[uc]);
			printk("base[%d]=%x,range[%d]=%x,result=%lx\n",uc,pada->base[uc],uc,pada->range[uc],pcam_ada->BaseAddress[uc]);
			//pcam_ada->BaseAddress[uc]&=0x000000FFFFFFFFFF;
		}
		else 
#endif
			pcam_ada->BaseAddress[uc] = (unsigned long)pada->base[uc];
	}

	/* enable bus master */
	pci_read_config_word(ulsata2_pci_dev,PCI_COMMAND,&command);
	command |= (PCI_COMMAND_MASTER | PCI_COMMAND_MEMORY);
	pci_write_config_word(ulsata2_pci_dev,PCI_COMMAND,command);
	pci_read_config_dword(ulsata2_pci_dev,0x30,&pada->rombase);

	/* register ioport */
#ifndef KERNEL_VERSION_25x
	if (!check_region(pada->base[0],pada->range[0])) {
		request_region(pada->base[0],pada->range[0],PDCNAME"(ATA)");
	}
	if (!check_region(pada->base[2],pada->range[2]))
		request_region(pada->base[2],pada->range[2],PDCNAME"(HOST)");
#else
	//if (!request_region(pada->base[0],pada->range[0],PDCNAME"(ATA)") ||
	//    !request_region(pada->base[2],pada->range[2],PDCNAME"(HOST)") ) {
	//	printk(KERN_WARNING "%s: Couldn't get io_region %#x-%#x.\n", PDCNAME, pada->base[0], pada->base[2]+pada->range[2]);
	//	return(1);}
	if (!request_region((unsigned long)pada->base[0],(unsigned long)pada->range[0],PDCNAME"(ATA)") ||
	    !request_region((unsigned long)pada->base[2],(unsigned long)pada->range[2],PDCNAME"(HOST)") ) {
		printk(KERN_WARNING "%s: Couldn't get io_region %#x-%#x.\n", PDCNAME, pada->base[0], pada->base[2]+pada->range[2]);
		return(1);}	
		
#endif
	
#ifndef KERNEL_VERSION_25x	
	if (request_irq(ulsata2_pci_dev->irq, ulsata2_interrupt, SA_INTERRUPT| SA_SHIRQ, PDCNAME,pada)) {
		printk("%s:[error] adapter%d request irq0x%x failed\n",PDCNAME, ada, ulsata2_pci_dev->irq);
		return(2);
	}
#endif	
	
	return(0);
}

/*
 * detect , initialize and register adapters.
 * return non-zero number on detection.
 */
int ulsata2_findcards(void) 
{
	unsigned char	class;
	unsigned char	adapter;
#ifdef	KERNEL_VERSION_24x
	struct pci_dev	*ulsata2_pci_dev = NULL ;
#else
	struct pci_dev	*ulsata2_pci_dev = pci_devices ;
#endif

	class = adapter = 0;
	while (UL_DEVID[class]) {
		/* search pci bus according to vendorID, deviceID */
		while( (ulsata2_pci_dev = pci_find_device(PCI_VENDOR_ID_PROMISE, UL_DEVID[class], ulsata2_pci_dev)) ) {
			DBG("vendor:device = %x:%x \n", PCI_VENDOR_ID_PROMISE, UL_DEVID[class] );
			if (pci_enable_device(ulsata2_pci_dev))
				continue;

#ifdef SCSI_DYNAMIC_DMA_MAPPING
			/* check dma capable */
			if (pci_set_dma_mask(ulsata2_pci_dev, 0xffffffff)) {
			       	printk(KERN_ERR "%s:[error] No suitable DMA available\n",PDCNAME);
				continue;
			}
#endif

			/* class code is 0x0180 or PDC579 */
			if(ulsata2_pci_dev->class>>8 == PCI_CLASS_STORAGE_OTHER || UL_DEVID[class] == DID579 || UL_DEVID[class] == DID779)
			{
				ulsata2_adapter_t  *pada = &ulsata2_adapter[adapter];
				unsigned char	code = 0;
				/* pdc579 class:0x0104 */
				/* offset 0x0d bit3 (0/1)=(FT/SATA) */
				if (UL_DEVID[class] == DID579 || UL_DEVID[class] == DID779) {
#if 0
					pci_read_config_byte(ulsata2_pci_dev, 0x0c, &code);
					if ( (code & 0x01) == 0 )
						continue;
#endif
#if 1
					pci_read_config_byte(ulsata2_pci_dev, 0x0d, &code);
					if ( (code & 0x08) == 0 )
						continue;
#endif
				}

				/* Check ASIC rev */
				pci_read_config_byte(ulsata2_pci_dev, 0x08, &pada->asic_rev);

				/* Fix Arima AMD64;
				Set Cache Line Size=1 and WMI=1 */
				code = 0x01 ;
				pci_write_config_byte(ulsata2_pci_dev, 0x0c, code);
				pci_read_config_byte(ulsata2_pci_dev, 0x42, &code);
				code |= 0x04 ;
				pci_write_config_byte(ulsata2_pci_dev, 0x42, code);

				/* over max cards */
				if (adapter >= MAXADAPTER ) {
				       	printk(KERN_WARNING "%s:[warning] max %d adapter(s) support\n",PDCNAME, MAXADAPTER);
					return(adapter);
				}
				//pada = &ulsata2_adapter[adapter];
				pada->present |= ulsata2_present;
				pada->pci_dev = ulsata2_pci_dev;
				pada->deviceid = ulsata2_pci_dev->device;
				pada->irq = ulsata2_pci_dev->irq;
				adapter++;
				DBG("irq = %#x\n", ulsata2_pci_dev->irq);
			}
		}
		class++;
	}
	return(adapter);
}

/*
 * start cam request packet
 */
ulsata2_req_t *ulsata2_request_start(void)
{
       	ulsata2_req_t *request, *nrequest;

	request = PoolH;
	nrequest = request->next;
	memset(request, 0, sizeof(ulsata2_req_t));
	PoolH = nrequest;

	return request;
}
/*
 * finish cam request packet
 */
void ulsata2_request_finish(ulsata2_req_t *request)
{
	PoolF->next = request;
	PoolF = request;
	PoolF->next = NULL;
	return;
}

void ulsata2_cbk_reset(unsigned char bch, unsigned char status)
{
	printk(KERN_INFO "%s:[info] Channel%2d reset status = %#x ",PDCNAME, bch, status);
}

/*
 * step down udma mode 
 */
void ulsata2_downmode(unsigned char drv)
{
	ulsata2_drive_t *pdrv;
	DEV_CONFIG *pcam_dev; 
	pdrv = &ulsata2_drive[drv];
	pcam_dev = pdrv->cam_dev;

	if (pcam_dev->UDMAMode != 0xFF) {
		if (pcam_dev->UDMAMode > 2) {
			unsigned char oldmode = pcam_dev->UDMAMode;
		       	pcam_dev->UDMAMode = (pcam_dev->UDMAMode-1 >= 2) ? (pcam_dev->UDMAMode-1): 2;
		       	if (CAM_ConfigATADevice(pcam_dev)) {
			       	CAM_ResetATAChannel(pcam_dev->bCh, ulsata2_cbk_reset);
			       	printk(KERN_INFO "%s:[info] disk%d downmode from %d to %d\n",PDCNAME,drv+1,oldmode,pcam_dev->UDMAMode);
			       	return;
		       	}
		}
	}
	return;
}

/*
 * Function:	ulsata2_decide_disposition()
 *
 * Purpose:	decode ATAStatus
 *
 * Parameters:	PATA_TASK	- pointer to ATA_TASK
 *
 * Return:	SCSI error code
 *
 * Notes:	n/a
 */
static int ulsata2_decide_disposition( PATA_TASK ptask )
{
	ulsata2_req_t *req = (ulsata2_req_t *) ptask->pTaskExtension;
       	unsigned short status;
       	//unsigned char status;
	int errcode = 0;
       	unsigned char id = 0;
	unsigned char ch = 0;
 	unsigned long lba = 0;
	ulsata2_drive_t *pdrv;

       	id = ptask->bCh * DrvPCh + ptask->bID;
       	ch = ptask->bCh;
	/* have to use backup LBAl */
	#ifdef _64BIT_
       	lba = (U32)(req->cam_task2.Cmd.ATACmd.LBAl);
       	#else
       	lba = req->cam_task2.Cmd.ATACmd.LBAl;
       	#endif
	pdrv = &ulsata2_drive[id];
	status = ptask->ATAStatus;
#if 0
	if ( status )
		printk("ATAStatus = %#x\n", status );
#endif
	if ( status & ATA_ABORT ) {
		printk(KERN_ERR "%s:[error] disk%d command abort at LBA %#lx cmd=%#x status=%#x error=%#x\n",PDCNAME,id+1,lba,ptask->Cmd.ATACmd.bCmd,ptask->Cmd.ATACmd.bStatus,ptask->Cmd.ATACmd.bError);
		errcode = DID_ABORT << 16;
		return(errcode);
	}

       	if (status & ATA_ERROR) {
		switch (status & ~ATA_ERROR) {
	        	case ATA_TIMEOUT: {
		       	    printk(KERN_WARNING "%s:[warning] disk%d ATA timeoutat LBA %#lx \n",PDCNAME,id+1,lba);
			    if (pdrv->cam_dev->DevFlag & DEV_EXIST) {
				    errcode = DID_BUS_BUSY << 16;
			    }
			    else {
					//gibsond
					//errcode = DID_TIME_OUT << 16;
					errcode = DID_ERROR << 16;
			    }
			    break;
		           }
			case ATAERR_OVERRUN: 
		       	    printk(KERN_ERR "%s:[error] disk%d ATA overrun\n",PDCNAME,id+1);
			    errcode = DID_ERROR << 16;
			    break;
			case ATAERR_UNDERRUN:
		       	    printk(KERN_ERR "%s:[error] disk%d ATA underrun\n",PDCNAME,id+1);
			    errcode = DID_ERROR << 16;
			    break;
			case ATAERR_PCI:
		       	    printk(KERN_ERR "%s:[error] PCI error\n",PDCNAME);
			    errcode = DID_ERROR << 16;
			    break;
			case ATAERR_DRV: {
			    /* ICRC bit on */
		       	    if (ptask->Cmd.ATACmd.bError & 0x80) {
				    printk(KERN_ERR "%s:[error] disk%d crc error at LBA %#lx\n",PDCNAME,id+1,lba);
				    ulsata2_downmode(id);
			    }
			    else {
				    printk(KERN_ERR "%s:[error] disk%d error at LBA %#lx cmd=%#x status=%#x error=%#x\n",PDCNAME,id+1,lba,ptask->Cmd.ATACmd.bCmd,ptask->Cmd.ATACmd.bStatus,ptask->Cmd.ATACmd.bError);
			    }
			    errcode = DID_ERROR << 16;
			    break;
			   }
			default: {
			    errcode = DID_ERROR << 16;
			   }
		}
       	}
	else 
		errcode = DID_OK << 16;

	return(errcode);
}

/*
 * submit ata callback function for ioctl
 */
void ulsata2_cbk_ioctl ( PATA_TASK ptask )
{
	ulsata2_req_t *req = (ulsata2_req_t *) ptask->pTaskExtension;
	unsigned char *argbuf = (unsigned char *)req->cam_task.DataBuffer;
	unsigned char *args = argbuf-8;

	unsigned char status;

	status = ptask->ATAStatus;

	if (status & ATA_ERROR) {
	       	ulsata2_decide_disposition(ptask);
       	}

	/* get register status */
	args[0] = ptask->Cmd.ATACmd.bFeature;
	args[1] = ptask->Cmd.ATACmd.bCount;
	args[2] = ptask->Cmd.ATACmd.LBAl & 0xFF;
	args[3] = ptask->Cmd.ATACmd.LBAl>>8 & 0xFF;
	args[4] = ptask->Cmd.ATACmd.LBAl>>16 & 0xFF;
	args[5] = ptask->Cmd.ATACmd.bCmd;
	args[6] = ptask->DataTransferLength >> 9;

	/* copy data to user */
	if (args[6]) {
	       	kfree(argbuf-8);
	}

	ulsata2_request_finish(req);

	ulsata2_unlock = 1;
	wake_up_interruptible(&wait);

	return;
}

/*********************************************************************
 * Function:
 *   update_swapbox		By Hank Yang
 *
 * Description:
 *   Update SwapBox info when timer active every 30 seconds.
*********************************************************************/
void update_swapbox(ulong device)
{
	unsigned char	sb;
	unsigned short	color;
	unsigned long	lock;
	//spinlock_t *host_lock;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,3)
	host_lock = &io_request_lock;
#else
	ulsata2_adapter_t *pada = &ulsata2_adapter[device];
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	spin_lock_irqsave(pada->ulsata2_lock, lock);
#else	
	spin_lock_irqsave(&CAM_spinlock, lock);
#endif	
	
	init_timer(&swapbox_timer);
	#ifndef	_64BIT_
	swapbox_timer.expires = jiffies + 30 * HZ ;
	#else
	swapbox_timer.expires = (unsigned long)jiffies + 30 * HZ ;
	#endif
	swapbox_timer.data = 0 ;
	swapbox_timer.function = update_swapbox ;
	add_timer(&swapbox_timer);

	for( sb=0 ; sb < MAXCHANNEL ; sb++ )
	{
		ENCLOSURE_TYPE	type;
		ENCLOSURE_STATUS box;
	       	unsigned char boxid;
		int	fail = 0;
	        //printk("\n SwapBox[%d]: Now Setting !!\n",sb); 
		if ( !((ulsata2_channel[sb].padapter->present)&ulsata2_present) )
			continue;
		if ( (boxid = CAM_EnumEnclosure(sb, &type)) == camNOBOX )
			continue;
		if (CAM_Get_Enclosure_Status(sb, &box) != camSUCCESS)
		       	continue;
		color = LED_Green ;
		if ( ((box.VOL5V/1000) < 4) || ((box.VOL5V/1000) >= 6) ) {
			//eventlog : 5V volt warning
			color = LED_Amber ;
		}
		if ( ((box.VOL12V/1000) < 11) || ((box.VOL12V/1000) >= 13) ) {
			//eventlog : 12V volt warning
			color = LED_Amber ;
		}
		//printk("\nCtrlLED[%d] to LED_Status=%x",sb,color);
		//printk("\nCtrlLED[%d] to LED_RED=%x",sb,color&LED_Red);
		//printk("\nCtrlLED[%d] to LED_Green=%x",sb,color&LED_Green);
		//printk("\nCtrlLED[%d] to LED_Amber=%x",sb,color&LED_Amber);
		//printk("\nCtrlLED[%d] to LEDON=%x\n",sb,color&LEDON);
	       	switch (type.BOXType) {
		       	case SuperSwap:
		       	case SuperSwap1100:
				if ( (box.TEMPStatus/2) > 52 )
					color = LED_Amber ;
				if ( box.FANStatus < 3000 )
					color = LED_Amber ;
				if ( color != p_LED[sb]) {
					SwapBox_CtrlLED(sb, color, LEDNOCHANGE, StatusLED );
				}
			       	break;
		       	case SuperSwap4100:
				if ( (box.TEMPStatus/2) > 52 ) {
					//eventlog : swapbox tempature too high
					color = LED_Amber ;
					fail = 1;
				}
				if ( (box.TEMP2Status/2) > 52) {
					//eventlog : swapbox tempature too high
					if ( fail )
						color = LED_Red ;
					else
						color = LED_Amber ;
				}
				if ( box.FANStatus < 3000 ) {
					//eventlog : fan rpm too low
					color = LED_Amber ;
					fail = 1;
				}
				if ( box.FAN2Status < 3000) {
					//eventlog : fan rpm too low
					if ( fail )
						color = LED_Red ;
					else
						color = LED_Amber ;
				}
				if ( color != p_LED[sb]) {
					SwapBox_CtrlLED(sb, color, LEDNOCHANGE, GlobalLED );
				}
			       	break;
	       	}
		p_LED[sb] = color ;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	spin_unlock_irqrestore(pada->ulsata2_lock, lock);
#else	
	spin_unlock_irqrestore(&CAM_spinlock, lock);
#endif	
	return ;
}

/*
 * initial device
 */
void ulsata2_init_device(unsigned long drv)
{
	ulsata2_drive_t *pdrv = &ulsata2_drive[drv];
	DEV_CONFIG *pcam_dev = pdrv->cam_dev;
	unsigned char bch = (char)drv / DrvPCh;
	unsigned char bid = (char)drv % DrvPCh;
	
	ulsata2_channel_t *pchn = &ulsata2_channel[bch];
	CHNL_CONFIG *pcam_chn  = pchn->cam_chn;
	
	/* prevent reinit channel */
	if ( bid == 0) {
		if (CAM_InitATAChannel(bch, NULL) != camSUCCESS) {
			//printk("CAM_InitATAChannel bch=%d Error!!!!\n\n",bch);
			return;
		}
		if (CAM_EnumATAChannel(pcam_chn) != camSUCCESS) {
			//printk("CAM_EnumATAChannel pcam_chn=%d Error!!!!!\n\n",pcam_chn);
			return;
		}
		DBG( "CAM_Init/EnumATAChannel( %d ) success.\n" , bch );
	}

	/* find device for each channel */
	memset(pcam_dev, 0, sizeof(DEV_CONFIG));
	pcam_dev->bCh = bch;
	pcam_dev->bID = bid;
	
	CAM_EnumATADevice(pcam_dev);
	
	if (pcam_dev->DevFlag & DEV_EXIST) {
		ENCLOSURE_TYPE	type;
		int		ret;

		/* check atapi is enable or disable */
		if (pcam_dev->DevFlag & DEV_ATAPI_DEVICE)
			if (!atapi)
				return;

		printk(KERN_INFO "%s:[info] Drive%2ld/%d: %s ",PDCNAME, drv+1,pcam_dev->bID, pcam_dev->Model);
		if (pcam_dev->MAXLBAL) {
			//int sz;
			unsigned long sz=0;
			
			/* scsi will add 1 in sd.c */
			pcam_dev->MAXLBAL -= 1;
			sz = (unsigned long)pcam_dev->MAXLBAL * 2;
			#ifdef _64BIT_
			printk("%10ds %6ldMB  ", pcam_dev->MAXLBAL, (sz/2 - sz/1250 + 974)/1950);
			#else
			printk("%10lds %6ldMB  ", pcam_dev->MAXLBAL, (sz/2 - sz/1250 + 974)/1950);
			#endif
		}
		else
			printk("%22s"," ");
		if (pcam_dev->UDMAMode != 0xFF)
			printk("UDMA%d\n", pcam_dev->UDMAMode);
		else if (pcam_dev->MDMAMode != 0xFF)
			printk("MDMA%d\n", pcam_dev->MDMAMode);
		else
			printk("PIO%d\n", pcam_dev->PIOMode);
#ifdef	SSBOX
		if ( ( ret = CAM_EnumEnclosure(bch, &type) ) != camNOBOX ) {
			printk("Enclosure(%x) detected.\n", bch);
			swapbox_exist = 1 ;
			SwapBox_CtrlLED(bch, LED_Green, LEDNOCHANGE, StatusLED );
			SwapBox_CtrlLED(bch, LEDON, LEDNOCHANGE, ActiveLED );
		}
#endif
	}
	return;
}

/*
 * initialize cam
 */
int ulsata2_init_cam(unsigned char ada)
{
	unsigned char uc,ud;
	unsigned char power = 0;
	unsigned long pages;
	unsigned long *cam_mem;
	ulsata2_adapter_t *pada = &ulsata2_adapter[ada];
	ADAPTER_CONFIG_INFO *pcam_ada = pada->cam_ada;
	CAM_INFO *pcam_info = pada->cam_info;
	ATA_MODULE_PACKET_COUNT pcnt;
	struct pci_dev	*ulsata2_pci_dev = NULL ;
	
	

	CAM_GetInfo(pcam_info);
	pcam_ada->Adapter_ID = pada->id;

	pages = (unsigned long)(pcam_info->MaxMemSize>>PAGE_SHIFT)+1;	/* in pages */
#ifdef NCQ
	pages = 16; //64K for NCQ
#endif
	while ( (1<<power) < pages ) power++;
	cam_mem = (unsigned long *) __get_dma_pages(GFP_ATOMIC, power);//K
	if ( cam_mem == NULL ) {
	       	printk(KERN_ERR "%s:[error] can't allocate enough memory for cam\n", PDCNAME);
		return(1);
	}
	pada->camVirMem = (void *)cam_mem;
	memset(cam_mem, 0, (1<<power)<<PAGE_SHIFT);
	pcam_ada->MemPhyAddress = virt_to_bus(cam_mem);//U64 = unsigned long
	pcam_ada->MemVirAddr = cam_mem;
	#ifdef _64BIT_
	pcam_ada->MemSize = (U32)((1<<power) << PAGE_SHIFT);
	#else
	pcam_ada->MemSize = (ulong) ((1<<power) << PAGE_SHIFT);
	#endif
	DBG("PhyAddr/VirAddr = %lx/%lx\n", (long)pcam_ada->MemPhyAddress, (long)pcam_ada->MemVirAddr);

#ifdef NCQ
	if( ( pcam_ada->MemPhyAddress & 0xFFFF0000 ) != ( ( pcam_ada->MemPhyAddress + MAX_CAM_MEMSIZE ) & 0xFFFF0000 ) ) {
		pcam_ada->MemPhyAddress = ((pcam_ada->MemPhyAddress + 0x0000FFFF) & 0xFFFF0000);
		pcam_ada->MemVirAddr = bus_to_virt( pcam_ada->MemPhyAddress );//void*=void*
		DBG("Adjusted NCQ memory boundary\n");
		DBG("PhyAddr/VirAddr = %lx/%lx\n", (long)pcam_ada->MemPhyAddress, (long)pcam_ada->MemVirAddr);
	}
#endif

	/* Check ASIC rev */
	if ( pada->asic_rev >= 2 )
		pcam_ada->TapeOutVersion = REVISION_2;
	else
		pcam_ada->TapeOutVersion = REVISION_1;
	
	printk("DEVICEID=%X\n",pada->deviceid);
	printk("Adapter_ID=%d, ada=%d\n",pcam_ada->Adapter_ID,ada);
	printk("BaseAddress = %lx\n",pcam_ada->BaseAddress[0]);
	printk("BaseAddress = %lx\n",pcam_ada->BaseAddress[1]);
	printk("BaseAddress = %lx\n",pcam_ada->BaseAddress[2]);
	printk("BaseAddress = %lx\n",pcam_ada->BaseAddress[3]);
	printk("BaseAddress = %lx\n",pcam_ada->BaseAddress[4]);
	printk("BaseAddress = %lx\n",pcam_ada->BaseAddress[5]);
	printk("Size = %ld\n",(unsigned long)pcam_ada->MemSize);
	printk("PhyAddr/VirAddr = %lx/%lx\n", (long)pcam_ada->MemPhyAddress, (long)pcam_ada->MemVirAddr);
	/* init cam */
	if (CAM_Init(pcam_ada) != camSUCCESS) {
	       	printk(KERN_ERR "%s:[error] cam init failed\n",PDCNAME);
		/* Add by Hank. 2003.09.17 */
	       	free_pages((ulong)pada->camVirMem, (unsigned int)power);
		return -1;
	}else
		printk("\n CAM_Init OK for adapter[%d]!!\n",ada);
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	ulsata2_pci_dev = pada->pci_dev;
	if (request_irq(ulsata2_pci_dev->irq, ulsata2_interrupt, SA_INTERRUPT | SA_SHIRQ, PDCNAME,pada)) {
		printk("%s:[error] adapter%d request irq0x%x failed\n",PDCNAME, ada, ulsata2_pci_dev->irq);
		return -1;
	}
	pada->present|=ulsata2_IN_INTR;
#endif
	if ( pada->asic_rev < 2 )
		CAM_ControlBMRLongBurst( ada, camDisable );
	else
		CAM_ControlBMRLongBurst( ada, camEnable );

	
#ifdef SSBOX
		CAM_InitSATAEnclosure(ada);
#endif

	/* init device */
	for (ud = 0; ud < DrvPCh*ChPAda; ud++) {
		unsigned long drv = (unsigned long)ada*DrvPCh*ChPAda + ud;
		ulsata2_init_device(drv);
	}

#ifdef SSBOX
	/* for SwapBox without hard drive */
	if (swapbox_exist){
		SwapBox_CtrlLEDEmptyDevices(ada, LED_Red, LEDNOCHANGE, StatusLED);
		SwapBox_CtrlLEDEmptyDevices(ada, LEDOFF, LEDNOCHANGE, StatusLED);

		SwapBox_CtrlLEDEmptyDevices(ada, LED_Green, LEDNOCHANGE, GlobalLED);
	}
#endif

       	/* get cam packet count */
       	CAM_ATAGetPacketCount(ada, &pcnt);

	/* set cam packet count to 1 to avoid overrun */
       	for (uc = 0; uc < ChPAda; uc++) {
	       	pcnt.ATAModule[uc] = 1;
       	}

       	/* set cam packet count */
       	CAM_ATASetPacketCount(ada, &pcnt);
	
	return(0);
}

/*
 * free resource: adapter dependent
 */
void ulsata2_cleanup(int ada)
{
	unsigned char i,j;
	ulsata2_adapter_t *pada = &ulsata2_adapter[ada];
	CAM_INFO *pcam_info = pada->cam_info;
	
	if (!((pada->present)&ulsata2_present))
	       	goto nextone;
	       	//return;

	
      
	/* free irq */
	free_irq(pada->pci_dev->irq,pada);
	 	
	 	
	 /* unregister ioport */
       	if (pada->base[0]) {
       		//#ifdef _64BIT_
	       	release_region((unsigned long)pada->base[0],(unsigned long)pada->range[0]);
	       	release_region((unsigned long)pada->base[2],(unsigned long)pada->range[2]);
	       	//#else
	       	//release_region(pada->base[0],pada->range[0]);
	       	//release_region(pada->base[2],pada->range[2]);
	       	//#endif
       	}
       	
	 	
	 	
	 	
	 	
	/* iounmap */
       	for (i = 0; i < 6; i++) {
#ifdef _MMIO_
	       	if (pcam_info->VirMemBase & bit(i) ) {
		       	iounmap((void *)pada->cam_ada->BaseAddress[i]);
	       	}
#endif
       	}
nextone:
	/* release cam memeory */
       	if (pada->cam_ada->MemSize) {
	       	unsigned char order = 0;
	       	while ( (1<<order) < (pada->cam_ada->MemSize>>PAGE_SHIFT) ) order++;
	       	free_pages((ulong)pada->camVirMem, (unsigned int)order);
       	}

	/* free memory */
	if (pada->cam_ada) {
		kfree(pada->cam_ada);
	}
	if (pada->cam_info) {
	       	kfree(pada->cam_info);
	}
	for (i = 0; i < ChPAda; i++) {
		if (pada->pchannel[i]->cam_chn) {
			kfree(pada->pchannel[i]->cam_chn);
		}
		for (j = 0; j < DrvPCh; j++) {
			if (pada->pchannel[i]->pdrive[j]->cam_dev) {
			       	kfree(pada->pchannel[i]->pdrive[j]->cam_dev);
			}
		}
	}
	return;
}
/*
 * free resource: adapter independent
 */
void ulsata2_cleanup2(void)
{
	unsigned char i;

	/* remove timer */
	for ( i = 0; i < MAXTIMER; i++) {
		if(ulsata2_timer[i].status != camFREE)
			del_timer(&ulsata2_timer[i].timer);
	}
	/* free queue memory */
	//for ( i = 0 ; i < QLENGTH +1; i++ ) {
	for ( i = 0 ; i < (QLENGTH*2)-1 ; i++ ) {
		if (ulsata2_req[i]) {
		       	kfree(ulsata2_req[i]);
		}
	}
	kfree(CamTimeoutTable);
	return;
}

/*
 * system shutdown routine
 */
static int ulsata2_halt(struct notifier_block *nb, unsigned long event, void *buf)
{
	switch (event) {
		case SYS_RESTART:
		case SYS_HALT:
		case SYS_POWER_OFF:
			break;
		default:
			return NOTIFY_DONE;
	}

	unregister_reboot_notifier(&ulsata2_notifier);

	return NOTIFY_DONE;
}

/* 
 * detect , initialize and register adapters.
 * return non-zero number on detection.
 */
int ulsata2_detect(Scsi_Host_Template *ulsata2_host) 
{
	unsigned char ua;
	unsigned char adapters;//, adapcnt;
	struct Scsi_Host *shpnt;

#ifdef	KERNEL_VERSION_24x
	struct pci_dev	*ulsata2_pci_dev = NULL ;
#else
	struct pci_dev	*ulsata2_pci_dev = pci_devices ;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,27)
       	ulsata2_host->proc_dir  = &proc_scsi_ulsata2;
#else
      	ulsata2_host->proc_name = PDCNAME;
#endif

	printk("PROMISE %s Series Linux Driver v%s\n", PRODUCT, VERSION);
	
	
	//printk("char=%d ,unsigned char=%d\n",sizeof(char),sizeof(unsigned char));
	//printk("short=%d ,unsigned short=%d\n",sizeof(short),sizeof(unsigned short));
	//printk("int=%d ,unsigned int=%d\n",sizeof(int),sizeof(unsigned int));
	//printk("long=%d ,unsigned long=%d long longs=%d\n",sizeof(long),sizeof(unsigned long),sizeof(long long));
	//printk("void=%d ,void *=%d\n",sizeof(void),sizeof(void*));
	//printk("char *=%d ,int *=%d\n",sizeof(char*),sizeof(int*));
	//printk("unsigned=%d ,signed=%d\n",sizeof(unsigned),sizeof(signed));
	//printk("SCSI_DATA_WRITE=%d\n",SCSI_DATA_WRITE);
	
	/* initialize global data structure */
	if(ulsata2_init_gds()) {
	       	printk(KERN_ERR "%s:[error] initialize global data structure failed\n",PDCNAME);
		return(0);
	}

	/* find ulsata2 */
	if ( !(adapters = ulsata2_findcards()) ) {
		printk(KERN_WARNING "%s:[warning] No valid cards found\n", PDCNAME);
		return(0);
	}

	/* register ioport */
	for(ua=0; ua < adapters; ua++) {
		if(!((ulsata2_adapter[ua].present)&ulsata2_present))
			continue;

	       	if (ulsata2_init_ioport(ua)) {
		       	printk(KERN_ERR "%s:[error] adapter%d: init ioport failed\n",PDCNAME,ua);
		       	return(0);
	       	}
	}

	/* initialize cam */
	for (ua = 0; ua < adapters; ua++) {
		if(!((ulsata2_adapter[ua].present)&ulsata2_present))
			continue;
        
		DBG("Initial adapter %d\n", ua);
	       	if (ulsata2_init_cam(ua)) {
		       	printk(KERN_ERR "%s:[error] adapter%d: init cam failed\n",PDCNAME,ua);
		       	return(0);
	       	}
	}
	
	

	
#ifdef SSBOX
	/* SwapBox Monitor Timer */
	if ( swapbox_exist ) {
		init_timer(&swapbox_timer);
		#ifndef	_64BIT_
		swapbox_timer.expires = jiffies + 30 * HZ ;
		#else
		swapbox_timer.expires = (unsigned long)jiffies + 30 * HZ ;
		#endif
		swapbox_timer.data = 0 ;
		swapbox_timer.function = update_swapbox ;
		add_timer(&swapbox_timer);
	}
#endif

	/* register scsi host */
	for (ua = 0; ua < adapters; ua++) {

	       	shpnt = scsi_register(ulsata2_host,sizeof(ulsata2_adapter_t));
	       	if (shpnt == NULL) {
		       	printk(KERN_ERR "%s:[error] adapter%d: register scsi host failed\n",PDCNAME,ua);
		       	ulsata2_cleanup(ua);
			continue;
	       	}
	       	shpnt->max_channel=0;
	       	shpnt->max_id=DrvPCh*ChPAda;
	       	shpnt->max_lun=1;


		
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	ulsata2_adapter_t *pada = &ulsata2_adapter[ua];
	ulsata2_pci_dev = pada->pci_dev;
	scsi_set_device(shpnt, &ulsata2_pci_dev->dev);	//scsi_set_device(struct Scsi_Host *shost,struct device *dev)?
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,4)
		scsi_set_pci_device(shpnt, ulsata2_pci_dev);
#endif

		
		if (shpnt->hostdata) {
			shpnt->irq = ulsata2_pci_dev->irq;
	       		pada = (ulsata2_adapter_t *) shpnt->hostdata;
	       		pada->host = shpnt;
			pada->present|=ulsata2_IN_INTR;
	 	      	memcpy(pada, &ulsata2_adapter[ua], sizeof(ulsata2_adapter_t));
			
		} else {
		       	printk(KERN_ERR "%s:[error] adapter%d: register scsi host failed\n",PDCNAME,ua);
		       	ulsata2_cleanup(ua);
			continue;
	       	}

       	}
       		

	ulsata2_num_adapters = adapters ;

	/* register some device */
	if (adapters) {
		/* register reboot notifier */
		if (register_reboot_notifier(&ulsata2_notifier)) {
			printk(KERN_WARNING "%s:[warning] shutdown routine register failed\n",PDCNAME); 
		}

		/* 
		 * register a character device for application to aceess 
		 * driver by ioctls.
		 */
		if ((ulsata2_major = register_chrdev(0, "ulsata2", 
						&ulsata2_fops)) < 0 ) {
			printk(KERN_ERR "%s:[error] register character device failed\n",PDCNAME);
		}

	}
	
	return(adapters);
}



/*
 * Ioctl Interface
 */
int ulsata2_ioctl(Scsi_Device *device, int cmd, void *arg)
{
	unsigned char drv;
	unsigned char *args;
	ulsata2_adapter_t *pada;
	ulsata2_drive_t *pdrv;
	
	pada =(ulsata2_adapter_t *)device->host->hostdata;
			
	args = (unsigned char *) arg;
					
	drv = device->id;
		
	pdrv = &ulsata2_drive[drv];
		
       	if( drv >= DrvPCh*ChPAda ) {
	       	//strcpy(args,"ioctl(): Drive not exist.");
	       	printk(KERN_DEBUG "ioctl(): Drive not exist.");
	       	return(-1);
       	}
       	if( !(pdrv->cam_dev->DevFlag & DEV_EXIST) ) {
       		//strcpy(args,"ioctl(): Drive not exist.");
	       	printk(KERN_DEBUG "ioctl(): Drive not exist.");
	       	return(-1);
       	}
	
	switch (cmd) {
	        /* RAW ATA CMD ** use carefully ** */
		case 0x3803: {
		     unsigned char *argbuf = args;
		     int argsize = 8, error=0;
		     ulsata2_req_t *queue;
		   
		     if ( !arg ) {
			     return(-1);
		     }
		   
		     /* support ATA device only */
		     if( !(pdrv->cam_dev->DevFlag & DEV_ATA_DEVICE) ) {
			     //strcpy(args,"ioctl(): ATA device only");
			     printk(KERN_DEBUG "ioctl(): ATA device only");
			     return(-1);
		     }
		   
		     /* there is data to trasnfer */
		     if (args[6]) {
			     argsize += (512 * args[6]);
			     argbuf = kmalloc(argsize, GFP_ATOMIC); //for Suse9.2
			     if (argbuf == NULL)
				     return(-1);
			     memset(argbuf, 0, argsize);
			     memcpy(argbuf, args, 8);
		     }
		
		     /* prepare cam task */
		     queue = ulsata2_request_start();
		     queue->cam_task.bCh=drv / DrvPCh;
		     queue->cam_task.bID=drv % DrvPCh;
		     queue->cam_task.ATACmdFlag=ATA_DEVICE+ATA_STATUS_RETURN;
		     if (args[6]) 
			     queue->cam_task.ATACmdFlag+=camDATA_XFER+PIO_XFER+camDATA_IN;
		     queue->cam_task.Cmd.ATACmd.bFeature = args[0];
		     queue->cam_task.Cmd.ATACmd.bCount=args[1];
		     queue->cam_task.Cmd.ATACmd.LBAl = args[2] | args[3] << 8| args[4] << 16;
		     queue->cam_task.Cmd.ATACmd.LBAh = 0;
		     queue->cam_task.Cmd.ATACmd.bCmd = args[5];
		     queue->cam_task.SGCount = 0;
		     queue->cam_task.DataTransferLength = args[6] << 9 ;
		     queue->cam_task.DataBuffer=argbuf+8;
		     queue->cam_task.callback = ulsata2_cbk_ioctl;
		     queue->buffer_size = argsize;
		     /* attach queue and backup cam_task */
		     queue->cam_task.pTaskExtension = (void *)queue;
		     memcpy(&queue->cam_task2, &queue->cam_task, 
				     sizeof(ATA_TASK));
		     
		     CAM_SubmitATA(&queue->cam_task);

		     ulsata2_unlock = 0;
		     error=wait_event_interruptible(wait, ulsata2_unlock);
			
		     if (args[6]) {
			     memcpy((void *)args, argbuf, argsize);
		     }
		     if(error)
		     	return(-1);
		     else		
		     	return(0) ;
		     }
		case SG_SET_TRANSFORM: {
		     return(0);
		     }
		default: {
		     if (debug & LOGATA) {
		       	 printk(KERN_WARNING "%s:[warning] unspported ioctl command(%#x)\n",PDCNAME,cmd); 
		     }
		     return(-1);
		     }
	}
}

/*
 * copy proc info, called by ulsata2_copy_info()
 */
static void ulsata2_copy_mem_info(ulsata2_info_t *info, char *data, int len)
{
	if (info->position + len > info->length)
		len = info->length - info->position;

	if (info->position + len < info->offset) {
		info->position += len;
		return;
	}
	if (info->position < info->offset) {
		data += (info->offset - info->position);
		len  -= (info->offset - info->position);
	}
	if (len > 0) {
		memcpy(info->buffer + info->position, data, len);
		info->position += len;
	}
}

/*
 * copy proc info, called by ulsata2_proc_info()
 */
static int ulsata2_copy_info(ulsata2_info_t *info, char *fmt, ...)
{
	va_list args;
	char buf[81];
	int len;

	va_start(args, fmt);
	len = vsprintf(buf, fmt, args);
	va_end(args);
	ulsata2_copy_mem_info(info, buf, len);
	return len;
}

/*
 * proc info
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
int ulsata2_proc_info(struct Scsi_Host *pHost, char *buffer, char **start, off_t offset, int length, int func) 
#else
int ulsata2_proc_info(char *buffer, char **start, off_t offset, int length, int hostno, int func)
#endif
{
	ulsata2_info_t info;
	unsigned char i;
	unsigned char SwapCount = 0;
	unsigned char disk = 0;
	int	SW4100_BoxID = 0;

	info.buffer = buffer;
	info.length = length;
	#ifdef _64BIT_
	info.offset = (int)offset;
	#else
	info.offset = offset;
	#endif
	info.position = 0;

	if (start)
		*start = buffer;

	ulsata2_copy_info(&info, "PROMISE %s Series Linux Driver v%s\n", PRODUCT, VERSION);

	/* get IC status */
	for (i = 0; i < MAXADAPTER; i++) {
		unsigned short id;
		if (!((ulsata2_adapter[i].present)& ulsata2_present))
			continue;
		id = ulsata2_adapter[i].deviceid;
		if ( id ) {
			ulsata2_copy_info(&info, "Adapter%d - ", i+1);
			switch (id) {
				case DID518:
					ulsata2_copy_info(&info, "SATAII 150 TX4  ");
					break;
				case DID575:
					ulsata2_copy_info(&info, "SATAII 150 TX2plus  ");
					break;
				case DID579:
					ulsata2_copy_info(&info, "SATAII 150 579  ");
					break;
				case DID718:
					ulsata2_copy_info(&info, "SATA 300 TX4  ");
					break;
				case DID775:
					ulsata2_copy_info(&info, "SATA 300 TX2plus  ");
					break;
				case DID779:
					ulsata2_copy_info(&info, "SATA 300 779  ");
					break;
				default:
					ulsata2_copy_info(&info, "SATA-II Series  ");
					break;
			}
			ulsata2_copy_info(&info, "[ ");
			ulsata2_copy_info(&info, "IRQ%2d ", ulsata2_adapter[i].irq);
			ulsata2_copy_info(&info, "]\n");
		}
	}

	/* get hard disk status */
	for (i = 0; i < MAXDRIVE; i++) {
		ulsata2_drive_t *pdrv = &ulsata2_drive[i];
		DEV_CONFIG *pcam_dev = pdrv->cam_dev;

		/* check drive exist or not */
		if (!(pcam_dev->DevFlag & DEV_EXIST))
			continue;

		/* check atapi is enable or disable */
	       	if (pcam_dev->DevFlag & DEV_ATAPI_DEVICE)
		       	if (!atapi)
			       	continue;

		if ( ++disk == 1 )
		       	ulsata2_copy_info(&info, "Drive    - \n");

		ulsata2_copy_info(&info, "  %2d : ", i+1);
		ulsata2_copy_info(&info, "%s ", pcam_dev->Model);
		ulsata2_copy_info(&info, "Channel%d/", pcam_dev->bCh%ChPAda+1);
		ulsata2_copy_info(&info, "%s ", (i%2)?"Slave ":"Master");
	       	if (pcam_dev->MAXLBAL) {
			//int sz = pcam_dev->MAXLBAL * 2;
			unsigned long sz = (unsigned long)pcam_dev->MAXLBAL * 2;
		       	ulsata2_copy_info(&info, "%6ldMB  ", (sz/2 - sz/1250 + 974)/1950);
		}
		else {
		       	ulsata2_copy_info(&info, "%10s", " ");
		}
		if (pcam_dev->UDMAMode != 0xFF)
		       	ulsata2_copy_info(&info, "UDMA%d\n", pcam_dev->UDMAMode);
		else if (pcam_dev->MDMAMode != 0xFF)
		       	ulsata2_copy_info(&info, "MDMA%d\n", pcam_dev->MDMAMode);
		else
		       	ulsata2_copy_info(&info, "PIO%d\n", pcam_dev->PIOMode);
	}

	/* get enclosure status */
	for (i = 0; i < MAXCHANNEL; i++) {
		ENCLOSURE_TYPE	type;
		ENCLOSURE_STATUS box;
		unsigned char boxid;
		int	SW4100_Flag = 0;
	        
		if ( !((ulsata2_channel[i].padapter->present)&ulsata2_present) )
			continue;
		if ( (boxid = CAM_EnumEnclosure(i, &type)) == camNOBOX )
			continue;
		if (CAM_Get_Enclosure_Status(i, &box) != camSUCCESS)
			continue;

		if ( !SwapCount )
			ulsata2_copy_info(&info, "SwapBox  - \n");

		switch (type.BOXType) {
	       	case SuperSwap:
				SwapCount++;
				ulsata2_copy_info(&info,"  %2d : SuperSwap      - ", SwapCount);
				SW4100_Flag = 0;
				break;
	       	case SuperSwap1100:
	       		SwapCount++;
				ulsata2_copy_info(&info,"  %2d : SuperSwap1100  - ", SwapCount);
				SW4100_Flag = 0;
				break;
	       	case SuperSwap4100:
#if 0
				if ( type.BOXID == 0 )
					continue;
#endif

				if (((type.BOXID & 0x0F) + 1) == SW4100_BoxID)
					continue;

				SwapCount++;
				SW4100_Flag	= 1;
				SW4100_BoxID = (type.BOXID & 0x0F) + 1;
				ulsata2_copy_info(&info,"  %2d : SuperSwap4100  - ", SwapCount);
				break;
	       	default:
		       	ulsata2_copy_info(&info,"unknow  - ");
				continue;
				break;
       	}

		if ( box.FANStatus < 1800 )
			box.FANStatus = 0;
		if ( box.FAN2Status < 1800 )
			box.FAN2Status = 0;

		if (SW4100_Flag){
			ulsata2_copy_info(&info,"RPM(%4ld/%4ld) ",box.FANStatus,box.FAN2Status);
			ulsata2_copy_info(&info,"TEMP(%d.",box.TEMPStatus/2);
			ulsata2_copy_info(&info,"%dC/",(box.TEMPStatus&1)*5);
			ulsata2_copy_info(&info,"%d.",box.TEMP2Status/2);
			ulsata2_copy_info(&info,"%dC) ",(box.TEMP2Status&1)*5);			
		}
		else{
			ulsata2_copy_info(&info,"RPM(%4ld) ",box.FANStatus);
			ulsata2_copy_info(&info,"TEMP(%d.",box.TEMPStatus/2);
			ulsata2_copy_info(&info,"%dC) ",(box.TEMPStatus&1)*5);
		}

/*
		if (SW4100_Flag)
			ulsata2_copy_info(&info,"RPM(%4ld/%4ld) ",box.FANStatus,box.FAN2Status);
		else
			ulsata2_copy_info(&info,"RPM(%4ld) ",box.FANStatus);
		
		ulsata2_copy_info(&info,"TEMP(%d.",box.TEMPStatus/2);
		
		if (SW4100_Flag) {
			ulsata2_copy_info(&info,"%dC/",(box.TEMPStatus&1)*5);
			ulsata2_copy_info(&info,"%d.",box.TEMP2Status/2);
			ulsata2_copy_info(&info,"%dC) ",(box.TEMP2Status&1)*5);
		}
		else
			ulsata2_copy_info(&info,"%dC) ",(box.TEMPStatus&1)*5);
*/

		ulsata2_copy_info(&info,"V5(%ld.",box.VOL5V/1000);
		ulsata2_copy_info(&info,"%02ldV) ",box.VOL5V%100);
		ulsata2_copy_info(&info,"V12(%ld.",box.VOL12V/1000);
		ulsata2_copy_info(&info,"%02ldV) ",box.VOL12V%100);
		ulsata2_copy_info(&info,"\n");
	}/* end of for (i = 0; i < MAXCHANNEL; i++)*/
	
	if (info.position > info.offset)
		return (info.position - info.offset);
	else
		return 0;
}

/*
 * transform SCSI CDB to ATAPI one
 */
static void ulsata2_scsi2atapi(ulsata2_req_t *req)
{
	Scsi_Cmnd *SCpnt = req->cmd;
	unsigned char *cmd = req->cam_task.Cmd.Cdb;
	unsigned char *scsi_buf = SCpnt->request_buffer;
	unsigned char *atapi_buf = NULL;

	switch (cmd[0]) {
		case READ_6:
		case WRITE_6:
			cmd[8] = cmd[4];
			cmd[5] = cmd[3];
			cmd[4] = cmd[2];
			cmd[3] = cmd[1] & 0x1F;
			cmd[2] = 0;
			cmd[1] &= 0xE0;
			cmd[0] += (READ_10 - READ_6);
			break;
		case MODE_SELECT:
		case MODE_SENSE:
			if ((cmd[0] == MODE_SELECT_10)||(cmd[0] == MODE_SENSE_10)) {
			if ((atapi_buf = kmalloc(req->buffer_size + 4, GFP_ATOMIC)) == NULL );//GFP_KERNEL)) == NULL )
				return;
				memset(atapi_buf, 0, req->buffer_size + 4);
			
			if ((cmd[0] == MODE_SELECT_10)) {
				atapi_buf[1] = scsi_buf[0];
				atapi_buf[2] = scsi_buf[1];
				atapi_buf[3] = scsi_buf[2];
				atapi_buf[7] = scsi_buf[3];
				memcpy(atapi_buf + 8, scsi_buf + 4, req->buffer_size - 4);
				req->buffer = atapi_buf;
							}
			}
			
			cmd[9] = cmd[5];
			cmd[8] = cmd[4] + 4;
			cmd[0] |= 0x40;
			if ( cmd[4] + 4 > 255)
				cmd[7] = cmd[4] + 4 - 255;
			cmd[4] = 0;
			cmd[5] = 0;
			
			req->buffer_size += 4;
			
			break;
		case 0x5b:	/* CLOSE TRACK/RZONE/SESSION/BORDER */
			cmd[1] |= 0x1;		/* set Immed bit */
			break;
		case 0xa1:	/* BLANK */
			cmd[1] |= 0x10;		/* set Immed bit */
			break;
		case 0xa6:	/* LOAD/UNLOAD MEDIUM */
			cmd[1] |= 0x1;		/* set Immed bit */
			break;
		case START_STOP: /* 0x1b */
			cmd[1] |= 0x1;		/* set Immed bit */
			break;
		case SYNCHRONIZE_CACHE:	/* 0x35 */
			cmd[1] |= 0x2;		/* set Immed bit */
			break;
		default:
			break;
	}
}
/*
 * transform ATAPI CDB to SCSI one
 */
static void ulsata2_atapi2scsi(ulsata2_req_t *req)
{
	Scsi_Cmnd *SCpnt = req->cmd;
	unsigned char *cmd = SCpnt->cmnd;
	unsigned char *scsi_buf = SCpnt->request_buffer;
	unsigned char *atapi_buf = req->buffer;

	switch (cmd[0]) {
		case INQUIRY:
			scsi_buf[2] |= 2;
			scsi_buf[3] = (scsi_buf[3] & 0xF0) | 2;
			break;
		case MODE_SENSE:
			if (req->cam_task.Cmd.Cdb[0] != MODE_SENSE_10) {
			if (atapi_buf && atapi_buf != scsi_buf)	
			       	{    		
			       	//if(ATAPI_BUF_Check)	//FIXME
				//	kfree(atapi_buf);
				//for Suse9.2
				}				
			       	break;
			}
			scsi_buf[0] = atapi_buf[1];
			scsi_buf[1] = atapi_buf[2];
			scsi_buf[2] = atapi_buf[3];
			scsi_buf[3] = atapi_buf[7];
			memcpy(scsi_buf + 4, atapi_buf + 8, req->buffer_size - 8);
			if (atapi_buf && atapi_buf != scsi_buf){
			//if(ATAPI_BUF_Check)
			//	kfree(atapi_buf);
					}
					
			break;
		default:
			break;
	}
}
/*
 * Function:	ulsata2_cbk_ata_rw()
 *
 * Purpose:	process cam task depend on ATAStatus
 *
 * Parameters:	PATA_TASK	- pointer to ATA_TASK
 *
 * Return:	nothing
 *
 * Notes:	n/a
 */
void ulsata2_cbk_ata_rw ( PATA_TASK ptask )
{
	
	ulsata2_req_t *req = (ulsata2_req_t *) ptask->pTaskExtension;
	Scsi_Cmnd *SCpnt = req->cmd;
#ifndef KERNEL_VERSION_25x
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->host->hostdata;
#else
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->device->host->hostdata;
#endif
	ulsata2_drive_t *pdrv;
	int result = DID_OK << 16;
	int id;

	id = (int)ptask->bCh * DrvPCh + ptask->bID;
	pdrv = &ulsata2_drive[id];

	/* debug */
	//if (debug & LOGATA) {
	if ((ptask->ATAStatus) || (req->cam_task.Cmd.ATACmd.bStatus)) {
		#ifdef _64BIT_
		printk(KERN_WARNING "%s:[%d] task done  retries=%d cmd=%#x flag=%#x lba=%#x cnt=%#x status=%#x e=%#x s=%#x\n", 
		#else
		printk(KERN_WARNING "%s:[%d] task done  retries=%d cmd=%#x flag=%#x lba=%#lx cnt=%#x status=%#x e=%#x s=%#x\n", 
		#endif
				PDCNAME,id,pdrv->retries,
				req->cam_task2.Cmd.ATACmd.bCmd,
				ptask->ATACmdFlag,
				req->cam_task2.Cmd.ATACmd.LBAl,
				req->cam_task2.Cmd.ATACmd.bCount,
				ptask->ATAStatus,
				req->cam_task.Cmd.ATACmd.bError,
				req->cam_task.Cmd.ATACmd.bStatus);
	} 

	unmap_scsi_data(pada, SCpnt);

	if (ptask->ATACmdFlag & camDATA_XFER) {
	       	result = ulsata2_decide_disposition(ptask);
	}

	/* if error then retry 5 times */
	//if ((result >> 16) != DID_OK) {
	if ( ((result >> 16) != DID_OK) && ((result >> 16) != DID_ABORT) ) {
		if (++pdrv->retries < 5) {
			ulsata2_request_finish(req);
			ulsata2_ata_rw(SCpnt);
			return;
		}
		/* retry too many times, abort it */
		else {
			if (CAM_ResetATAChannel(ptask->bCh, ulsata2_cbk_reset) != camSUCCESS) {
				printk(KERN_WARNING "%s:[error] reset channel %d failed\n",PDCNAME, ptask->bCh);
			}

			result = DID_ABORT<<16;
		}
	}

	pdrv->retries = 0;
	SCpnt->result = result;
	SCpnt->scsi_done(SCpnt);
	ulsata2_request_finish(req);

	return;
}
/*
 * SubmitATA callback for ulsata2_atapi_dma
 */
void ulsata2_cbk_atapi_dma( PATA_TASK ptask )
{
	ulsata2_req_t *req = (ulsata2_req_t *) ptask->pTaskExtension;
	Scsi_Cmnd *SCpnt = req->cmd;
#ifndef KERNEL_VERSION_25x
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->host->hostdata;
#else
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->device->host->hostdata;
#endif

	/* unmap DMA mapping */
	if (SCpnt->use_sg)
	       	unmap_scsi_data(pada, SCpnt);
	else {
	       	int dma_dir; 
		if (__data_mapped(SCpnt) == 1) {
		       	dma_dir = scsi_to_pci_dma_dir(SCpnt->sc_data_direction);
		       	pci_unmap_single(pada->pci_dev, __data_mapping(SCpnt), (size_t)req->buffer_size, dma_dir);
		       	//pci_unmap_single(struct pci_dev *pdev, dma_addr_t bus_addr, size_t sz, int direction)
		       	__data_mapped(SCpnt) = 0;
		}
	}

	ulsata2_atapi2scsi(req);
	if (debug & LOGATAPI) {
	       	printk("    status=%2x   ",ptask->ATAStatus);
	       	if (SCpnt->request_bufflen) {
		       	printk("rst=");
		       	hexdump(SCpnt->request_buffer, minmax(16, SCpnt->request_bufflen));
	       	}
	       	else
		       	printk("\n");
       	}

	/* check ATAStatus */
	if (ptask->ATAStatus & ATA_ERROR) {
	       	if (ptask->ATAStatus & ATA_TIMEOUT)
		       	SCpnt->result = (DID_BUS_BUSY << 16);
	       	else if (ptask->ATAStatus & ATAERR_DRV) {
		       	SCpnt->result = (CHECK_CONDITION << 1) | (DID_OK << 16);
			if (debug & LOGATAPI) {
			       	printk("              sense=");
			       	hexdump(SCpnt->sense_buffer, 14);
			}
	       	}
	}
	else {
	       	SCpnt->result = (DID_OK << 16);
	}

	SCpnt->scsi_done(SCpnt);
	ulsata2_request_finish(req);

	return;
}
/*
 * SubmitATA callback for ulsata2_atapi_pio
 */
void ulsata2_cbk_atapi_pio( PATA_TASK ptask )
{
	ulsata2_req_t *req = (ulsata2_req_t *) ptask->pTaskExtension;
	Scsi_Cmnd *SCpnt = req->cmd;
#ifndef KERNEL_VERSION_25x
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->host->hostdata;
#else
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->device->host->hostdata;
#endif
	
	ulsata2_atapi2scsi(req);

	if (debug & LOGATAPI) {
	printk("    status=%2x   ",ptask->ATAStatus);
	printk("rst=");
       	hexdump(req->buffer, minmax(16, SCpnt->request_bufflen));
	}

       	if (SCpnt->use_sg) {
       		if (SCpnt->sc_data_direction != SCSI_DATA_WRITE) {
	       	       	unsigned char *s, *t;
		       	struct scatterlist *sglist;
		       	int i, sgcnt, addr=0;

			sglist = (struct scatterlist *)SCpnt->request_buffer;
		       	sgcnt = (int)SCpnt->use_sg;
		       	for (i = 0; i < sgcnt; i++) {
			       	unsigned int len = scsi_sg_dma_len(sglist+i);
			       	s = &req->buffer[addr];
			       	t = bus_to_virt(scsi_sg_dma_address(sglist+i));
			       	memcpy(t, s, len);
			       	addr += len;
		       	}
		       	
		}
		__data_mapped(SCpnt) = 2;
		#ifdef _64BIT_
		__data_mapping(SCpnt) = (dma_addr_t)SCpnt->use_sg;
		#else
		__data_mapping(SCpnt) = SCpnt->use_sg;
		#endif
		unmap_scsi_data(pada, SCpnt);
		//if((ATAPI_BUF_Check!=1)&&(req->buffer))	//Only for Suse9.2  //FIXME!!!!!!UNMAKE SENSE!!!!!
			kfree(req->buffer);		
		//ATAPI_BUF_Check=0;		   	        //for Suse9.2
	}

	if (ptask->ATAStatus & ATA_ERROR) {
	       	if (ptask->ATAStatus & ATA_TIMEOUT)
		       	SCpnt->result = (DID_BUS_BUSY << 16);
	       	else if (ptask->ATAStatus & ATAERR_DRV) {
		       	SCpnt->result = (CHECK_CONDITION << 1) | (DID_OK << 16);
			if (debug & LOGATAPI) {
			       	printk("              sense=");
			       	hexdump(SCpnt->sense_buffer, 14);
			}
	       	}
	}
	else
	       	SCpnt->result = (DID_OK << 16);
	SCpnt->scsi_done(SCpnt);
	
	ulsata2_request_finish(req);
}
/*
 * Function:	ulsata2_ata_rw()
 *
 * Purpose:	process scsi R/W commands and submit it to CAM
 *
 * Parameters:	SCpnt	- scsi command structure
 *
 * Return:	0 is success, 1 is fail
 *
 * Notes:	n/a
 */
static int ulsata2_ata_rw(Scsi_Cmnd *SCpnt)
{
	//unsigned char id;
	unsigned int id;
	ulsata2_req_t *queue;
	unsigned char *cmd = (unsigned char *) SCpnt->cmnd;
	
#ifndef KERNEL_VERSION_25x
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->host->hostdata;
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->target;
#else
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->device->host->hostdata;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->id;
#else
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->channel;
#endif	
#endif

	queue = ulsata2_request_start();

	/* attach SCpnt for DMA mapping */
	queue->cmd = SCpnt;
	queue->cam_task.bCh = (U8)(id / DrvPCh);
	queue->cam_task.bID = (U8)(id % DrvPCh);

       	queue->cam_task.ATACmdFlag = ATA_DEVICE+camDATA_XFER+PSEUDO_CMD+PSEUDO_TAG_CMD;

	/* assign bCount and LBA depend on scsi cmd */
	switch (cmd[0]) {
		case READ_6:
		       	queue->cam_task.ATACmdFlag += camDATA_IN;
		       	queue->cam_task.Cmd.ATACmd.bCmd = IDE_COMMAND_READ_DMA;
		       	queue->cam_task.Cmd.ATACmd.bCount = cmd[4];
		       	queue->cam_task.Cmd.ATACmd.LBAl = ( ((cmd[1]&0x1F) << 16) | (cmd[2] << 8) | cmd[3]);
			break;
		case READ_10:
		       	queue->cam_task.ATACmdFlag += camDATA_IN;
		       	queue->cam_task.Cmd.ATACmd.bCmd = IDE_COMMAND_READ_DMA;
		       	queue->cam_task.Cmd.ATACmd.bCount = ( (cmd[7] << 8) | cmd[8]);
		       	queue->cam_task.Cmd.ATACmd.LBAl = ( (cmd[2] << 24) | (cmd[3] << 16) | (cmd[4] << 8) | cmd[5]);
			break;
		case WRITE_6:
		       	queue->cam_task.Cmd.ATACmd.bCmd = IDE_COMMAND_WRITE_DMA;
		       	queue->cam_task.Cmd.ATACmd.bCount = cmd[4];
		       	queue->cam_task.Cmd.ATACmd.LBAl = ( ((cmd[1]&0x1F) << 16) | (cmd[2] << 8) | cmd[3]);
			break;
		case WRITE_10:
		       	queue->cam_task.Cmd.ATACmd.bCmd = IDE_COMMAND_WRITE_DMA;
		       	queue->cam_task.Cmd.ATACmd.bCount = ( (cmd[7] << 8) | cmd[8]);
		       	queue->cam_task.Cmd.ATACmd.LBAl = ( (cmd[2] << 24) | (cmd[3] << 16) | (cmd[4] << 8) | cmd[5]);
			break;
	}

	queue->cam_task.Cmd.ATACmd.LBAh = 0;
	queue->cam_task.callback = ulsata2_cbk_ata_rw;

	/* there is one sg table, request_buffer point to sg list */
	if (SCpnt->use_sg) {
		struct scatterlist *sglist;
		int i, sgcnt;

		sglist = (struct scatterlist *)SCpnt->request_buffer;
		sgcnt = map_scsi_sg_data(pada, SCpnt);

		for (i = 0; i < sgcnt; i++) {
		       	unsigned int len = scsi_sg_dma_len(sglist+i);
		       	#ifdef _64BIT_
			queue->cam_sg[i].StartAddr = (U32)cpu_to_le32(scsi_sg_dma_address(sglist+i));
			#else
			queue->cam_sg[i].StartAddr = cpu_to_le32(scsi_sg_dma_address(sglist+i));
			#endif
			queue->cam_sg[i].wCount = (U16) len;
	       	}
	       	queue->cam_sg[sgcnt-1].CtrlFlag = EOT;
		queue->cam_task.SGCount = sgcnt;
		queue->cam_task.pSG = (unsigned long)((PCAMSG)&queue->cam_sg[0]);
	}
	/* request_buffer is a virtual addr point to real sg address , 
	 * convert it into our sg table
	 */
	else {
		dma_addr_t address;
		PCAMSG psg;
		address = map_scsi_single_data(pada, SCpnt);
		psg = &queue->cam_sg[0];
		#ifdef _64BIT_
		psg->StartAddr = (U32)cpu_to_le32(address);
		#else
		psg->StartAddr = cpu_to_le32(address);
		#endif
		psg->wCount = (U16) SCpnt->request_bufflen;
		psg->Reserved = 0;
		psg->CtrlFlag = EOT;
		queue->cam_task.SGCount = 1;
		queue->cam_task.pSG = (unsigned long)((PCAMSG)psg);
	}

	/* attach queue structure */
	queue->cam_task.pTaskExtension = (void *)queue;

	/* backup cam_task */
	memcpy(&queue->cam_task2, &queue->cam_task, sizeof(ATA_TASK));

	/* debug */
	if (debug & LOGATA) {
		#ifdef _64BIT_
		printk(KERN_WARNING "%s:[%d] task start cmd=%#x lba=%#x cnt=%#x\n", 
		#else
		printk(KERN_WARNING "%s:[%d] task start cmd=%#x lba=%#lx cnt=%#x\n", 
		#endif
				PDCNAME,id,
				queue->cam_task.Cmd.ATACmd.bCmd,
				queue->cam_task.Cmd.ATACmd.LBAl,
				queue->cam_task.Cmd.ATACmd.bCount
				);
	}

	if (CAM_SubmitATA(&queue->cam_task) != camACCEPTED) {
		ulsata2_request_finish( queue );
		printk(KERN_WARNING "%s:[warning] submit channel%2d busy\n",PDCNAME, queue->cam_task.bCh );
		return(1);
	}
	return(0);
}

/*
 * ATAPI CDB DMA mode
 */
static int ulsata2_atapi_dma(Scsi_Cmnd *SCpnt)
{
	unsigned char *buff = (unsigned char *) SCpnt->request_buffer;
	//unsigned char id;
	unsigned int id;
	int bufflen = (int)SCpnt->request_bufflen;
       	PCAMSG psg;
	ulsata2_req_t *queue;
#ifndef KERNEL_VERSION_25x
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->host->hostdata;
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->target;
#else
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->device->host->hostdata;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->id;
#else
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->channel;
#endif		
#endif

	queue = ulsata2_request_start();
       	queue->cmd = SCpnt;
       	queue->buffer=buff;
       	queue->buffer_size=bufflen;
       	queue->cam_task.bCh=(U8)(id / DrvPCh);
       	queue->cam_task.bID=(U8)(id % DrvPCh);
       	queue->cam_task.ATACmdFlag=0;
       	memcpy(queue->cam_task.Cmd.Cdb, SCpnt->cmnd, SCpnt->cmd_len);
       	ulsata2_scsi2atapi(queue);
       	if (SCpnt->use_sg) {
	       	struct scatterlist *sglist;
	       	int i, sgcnt;

		sglist = (struct scatterlist *)SCpnt->request_buffer;
	       	sgcnt = map_scsi_sg_data(pada, SCpnt);
		for (i = 0; i < sgcnt; i++) {
		       	unsigned int len = scsi_sg_dma_len(sglist+i);
			#ifdef	_64BIT_
			queue->cam_sg[i].StartAddr = (U32)(cpu_to_le32(scsi_sg_dma_address(sglist+i)));
			#else
			queue->cam_sg[i].StartAddr = cpu_to_le32(scsi_sg_dma_address(sglist+i));
			#endif
			queue->cam_sg[i].wCount = cpu_to_le32(len);
	       	}
	       	queue->cam_sg[sgcnt-1].CtrlFlag = EOT;

	       	queue->cam_task.SGCount = sgcnt;
	       	queue->cam_task.pSG = (unsigned long)((PCAMSG)&queue->cam_sg[0]);
       	}
       	else { /* use_sg = 0 */
	       	if (queue->buffer_size) {
		       	dma_addr_t address;
		       	int dma_dir = scsi_to_pci_dma_dir(SCpnt->sc_data_direction);
		       	address = pci_map_single(pada->pci_dev, queue->buffer, queue->buffer_size, dma_dir);
		       	__data_mapped(SCpnt) = 1;
		       	__data_mapping(SCpnt) = address;

			psg = &queue->cam_sg[0];
			#ifdef	_64BIT_
			psg->StartAddr = (U32)cpu_to_le32(address);
			psg->wCount = (U16)cpu_to_le32(queue->buffer_size);
			#else
		       	psg->StartAddr = cpu_to_le32(address);
		       	psg->wCount = cpu_to_le32(queue->buffer_size);
		       	#endif
		       	psg->Reserved = 0;
		       	psg->CtrlFlag = EOT;
		       	queue->cam_task.SGCount = 1;
		       	queue->cam_task.pSG = (unsigned long)((PCAMSG)psg);
	       	}
       	}
       	if (queue->cam_task.SGCount) {
	       	queue->cam_task.ATACmdFlag+=camDATA_XFER;
	       	if (SCpnt->sc_data_direction != SCSI_DATA_WRITE)
		       	queue->cam_task.ATACmdFlag+=camDATA_IN;
       	}
       	queue->cam_task.callback = ulsata2_cbk_atapi_dma;
       	queue->cam_task.SenseInfoLength = 14;
       	//queue->cam_task.SenseInfoLength = 0;	//DEBUG
       	queue->cam_task.SenseInfoBuffer = (void *)SCpnt->sense_buffer; //void * = char *

	/* attach queue structure */
	queue->cam_task.pTaskExtension = (void *)queue;

	/* backup cam_task */
	memcpy(&queue->cam_task2, &queue->cam_task, sizeof(ATA_TASK));

	if (CAM_SubmitATA(&queue->cam_task) != camACCEPTED) {
		printk(KERN_WARNING "%s:[warning] submit channel%2d busy\n",PDCNAME, queue->cam_task.bCh );
		return(1);
	}

	if (debug & LOGATAPI) {
	       	printk("dma: cmd=%x flag=%x len=%x sg=%x dir=%x\n",SCpnt->cmnd[0],queue->cam_task.ATACmdFlag,bufflen,SCpnt->use_sg,SCpnt->sc_data_direction);
	       	printk("%s: sg=%d cmd=", PDCNAME, SCpnt->use_sg);
	       	hexdump(queue->cam_task.Cmd.Cdb, SCpnt->cmd_len);
	}
	return(0);
}

/*
 * ATAPI CDB PIO mode
 */
static int ulsata2_atapi_pio(Scsi_Cmnd *SCpnt)
{
	unsigned char *buff = (unsigned char *) SCpnt->request_buffer;
	//unsigned char id;
	unsigned int id;
	int bufflen = (int)SCpnt->request_bufflen;
	ulsata2_req_t *queue;
#ifndef KERNEL_VERSION_25x
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->host->hostdata;
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->target;
#else
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->device->host->hostdata;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->id; //U8=U8+U16?
#else
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->channel;//U8=U8+U16?
#endif	
#endif
	
       	queue = ulsata2_request_start();
       	queue->cmd = SCpnt;
       	queue->buffer=buff;
       	queue->buffer_size=bufflen;
       	queue->cam_task.bCh=(U8)(id / DrvPCh);
       	queue->cam_task.bID=(U8)(id % DrvPCh);
       	queue->cam_task.ATACmdFlag=0;
       	memcpy(queue->cam_task.Cmd.Cdb, SCpnt->cmnd, SCpnt->cmd_len);
       	ulsata2_scsi2atapi(queue);
       	if (SCpnt->use_sg) {
       		if ((queue->buffer = kmalloc(bufflen, GFP_ATOMIC)) == NULL ) { //GFP_KERNEL))==NULL){//for Suse9.2
		       	printk(KERN_ERR "%s:[error] can't allocate enough memory for sg\n",PDCNAME);
		       	ulsata2_request_finish(queue);
			return(1);
	       	}
	       	
	       	memset(queue->buffer, 0, bufflen);
	       	if (SCpnt->sc_data_direction == SCSI_DATA_WRITE) {
		       	unsigned char *s, *t;
		       	struct scatterlist *sglist;
		       	int i, sgcnt, addr=0;

			sglist = (struct scatterlist *)SCpnt->request_buffer;
		       	sgcnt = (int)SCpnt->use_sg;//int=unsign short
		       	for (i = 0; i < sgcnt; i++) {
			       	unsigned int len = scsi_sg_dma_len(sglist+i);
			       	s = bus_to_virt(scsi_sg_dma_address(sglist+i));
			       	t = &queue->buffer[addr];
			       	memcpy(t, s, len);
			       	addr += len;
		       	}
	       	}
	}
       	if (queue->buffer_size) {
	       	queue->cam_task.ATACmdFlag+=PIO_XFER;
	       	queue->cam_task.ATACmdFlag+=camDATA_XFER;
	       	if (SCpnt->sc_data_direction != SCSI_DATA_WRITE)
		       	queue->cam_task.ATACmdFlag+=camDATA_IN;
       	}
       	queue->cam_task.SGCount = 0;
       	queue->cam_task.DataBuffer=(void *)queue->buffer;
       	queue->cam_task.DataTransferLength=(U32)queue->buffer_size;
       	queue->cam_task.SenseInfoLength = 14;
       	queue->cam_task.SenseInfoBuffer = (void *)SCpnt->sense_buffer;
       	queue->cam_task.callback=ulsata2_cbk_atapi_pio;

	/* attach queue structure */
	queue->cam_task.pTaskExtension = (void *)queue;

	/* backup cam_task */
	memcpy(&queue->cam_task2, &queue->cam_task, sizeof(ATA_TASK));

	if (CAM_SubmitATA(&queue->cam_task) != camACCEPTED) {
		printk(KERN_WARNING "%s:[warning] submit channel%2d busy\n",PDCNAME, queue->cam_task.bCh );
		return(1);
	}

	if (debug & LOGATAPI) {
	       	printk("pio: cmd=%x flag=%x len=%x sg=%x dir=%x\n",SCpnt->cmnd[0],queue->cam_task.ATACmdFlag,bufflen,SCpnt->use_sg,SCpnt->sc_data_direction);
	       	printk("%s: sg=%d cmd=", PDCNAME, SCpnt->use_sg);
	       	hexdump(SCpnt->cmnd, SCpnt->cmd_len);
	}
	return(0);
}
/*
 * Function:	ulsata2_queuecommand()
 *
 * Purpose:	scsi queue function to handle scsi commands
 *
 * Parameters:	SCpnt	- scsi command structure
 * 		done	- scsi done callback 
 *
 * Return:	only when use_new_eh_code = 1, return value is meaningful
 *
 * Notes:	n/a
 */
int ulsata2_queuecommand(Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *))
{
	unsigned char *cmd = (unsigned char *) SCpnt->cmnd;
	unsigned char *buff = (unsigned char *) SCpnt->request_buffer;
	//unsigned char drv = 0;
	unsigned int drv = 0;
	unsigned long lock = 0;
	unsigned long capacity;
	int bufflen = (int)SCpnt->request_bufflen;
	int error = 0;
	//unsigned char id;
	unsigned int id;
	static unsigned int smartstatus = 0 ;
	ulsata2_adapter_t *pada;
	ulsata2_drive_t *pdrv;
	DEV_CONFIG *pcam_dev;
	ulsata2_req_t *queue;
	//MODULE_TIMEOUT CamTimeoutTable;
	
	__data_mapped(SCpnt) = 0;
	__data_mapping(SCpnt) = 0;

	if ( debug ) {
		printk("%s:[D] Qcmd t=%d cmd=",
#ifndef KERNEL_VERSION_25x
				PDCNAME, SCpnt->target);
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
			PDCNAME, SCpnt->device->id);
#else
			PDCNAME, SCpnt->device->channel);
#endif	
				
#endif
		hexdump(SCpnt->cmnd, SCpnt->cmd_len);
	if(SCpnt == NULL)
		printk("%s:[error] invalid point in SCpnt\n",PDCNAME);
	}

#ifndef KERNEL_VERSION_25x
	pada = (ulsata2_adapter_t *) SCpnt->host->hostdata;
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->target;
#else
	pada = (ulsata2_adapter_t *) SCpnt->device->host->hostdata;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->id;
#else
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->channel;
#endif	
#endif
	//printk("id=%d\n",id);
	if (pada == NULL ) {
	       	printk(KERN_ERR "%s:[error] invalid device in queue cmd\n",PDCNAME);
		SCpnt->result = (DID_ERROR << 16);
		done(SCpnt);
		return(0);
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	//printk("present=%d ulsata2_lock=%p &CAM_spinlock=%lx\n",pada->present,pada->ulsata2_lock,&CAM_spinlock);
	spin_lock_irqsave(pada->ulsata2_lock, lock);
#else
	spin_lock_irqsave(&CAM_spinlock, lock);
#endif	
	
	
#ifndef KERNEL_VERSION_25x
	drv = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->target;
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	drv = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->id;
#else
	drv = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->channel;
#endif
#endif
	pdrv = &ulsata2_drive[drv];
	pcam_dev = pdrv->cam_dev;
	
	/* drive not exist */
//	if (!(pcam_dev->DevFlag & DEV_EXIST)) {
	if ( pcam_dev->DevFlag == DEV_NOTFOUND ) {
		//printk("DEV_NOTFOUND drv=%d id=%d/%d\n",drv,id,pada->id);
       		SCpnt->result = DID_BAD_TARGET << 16 ;
		done(SCpnt);
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		spin_unlock_irqrestore(pada->ulsata2_lock, lock);
	#else	
		spin_unlock_irqrestore(&CAM_spinlock, lock);
	#endif	
		return(0);
	}

	/* check atapi is enable or disable */
       	if (pcam_dev->DevFlag & DEV_ATAPI_DEVICE) {
	       	if (!atapi) {
		       	SCpnt->result = DID_BAD_TARGET << 16 ;
		       	done(SCpnt);
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)	       	
		       	spin_unlock_irqrestore(pada->ulsata2_lock, lock);
	#else	       	
		       	spin_unlock_irqrestore(&CAM_spinlock, lock);
	#endif	       	
		       	return(0);
		}
	}

	SCpnt->scsi_done = done;
	
	/* ATAPI device */
	if (pcam_dev->DevFlag & DEV_ATAPI_DEVICE) {
		unsigned char pio = 1;

		/* use dma to do data transfer */
		switch (cmd[0]) {
			case WRITE_12:
			case WRITE_10:
			case WRITE_6:
			case READ_12:
			case READ_10:
			case READ_6:
			case 0xad:
			case 0xbe:
		//	case 0x03:	//DEBUG
				pio = 0; break;
		}

		/* -45150 (FFFF4FA2h) to -1 (FFFFFFFFh) shall use PIO mode */
		if (cmd[0] == WRITE_10) {
			unsigned long lba;
			lba = cmd[5] | cmd[4] << 8 | cmd[3] << 16| cmd[2] << 24;
			if (lba >= 0xFFFF4FA2 && lba <= 0xFFFFFFFF) {
				pio = 1;
			}
		}

		/* device support only pio mode */
	       	if (pcam_dev->UDMAMode == 0xFF && pcam_dev->MDMAMode == 0xFF) {
			pio = 1;
	       		DBG(KERN_WARNING "%s:[warning] ATAPI device do not support pio mode.\n",PDCNAME);
		}

		/* reset CAM timeout timer for each atapi task */
		//memset(&CamTimeoutTable, 0, sizeof(CamTimeoutTable)); //for Suse9.2
		//CamTimeoutTable.ATAPICommand = SCpnt->timeout_per_command * 18 / HZ;
		//CAM_SetTimeoutTime((U8)pada->id, &CamTimeoutTable);
		memset(CamTimeoutTable, 0, sizeof(MODULE_TIMEOUT));
		CamTimeoutTable->ATAPICommand = (U32)SCpnt->timeout_per_command * 18 / HZ;
		CAM_SetTimeoutTime((U8)pada->id, CamTimeoutTable);
		
		/* atapi dma mode */
		if (!pio) {
			error = ulsata2_atapi_dma(SCpnt);
		}
		else { /* atapi pio mode */
		       	error = ulsata2_atapi_pio(SCpnt);
		}

	}
	else { /* ATA device */
		switch (cmd[0]) {
			case INQUIRY:
				{
				memset(buff,0 , bufflen);
				buff[0] = TYPE_DISK;	/* device type */
				buff[1] = 0;		/* RMB and modifier (SCSI-1) */
				buff[2] = 2;		/* claim SCSI 2 */
				buff[4] = 31;		/* Additional len (n-4) */
				memcpy(&buff[16], pcam_dev->Model, 16);
				memcpy(&buff[32], "    ", 4);
				SCpnt->result = (DID_OK << 16);
				done(SCpnt);
				}
				break;
			case TEST_UNIT_READY:
				SCpnt->result = (DID_OK << 16);
				done(SCpnt);
				break;
			case READ_CAPACITY:
				{
				capacity = (unsigned long)pcam_dev->MAXLBAL;
				memset(buff,0 , bufflen);
				/* Number of LBA's */
				buff[0] = (unsigned char) (capacity >> 24);
				buff[1] = (unsigned char) (capacity >> 16) & 0xff;
				buff[2] = (unsigned char) (capacity >> 8) & 0xff;
				buff[3] = (unsigned char) capacity & 0xff;
				/* Block size in bytes (512) */
				buff[4] = 0;
				buff[5] = 0;
				buff[6] = 0x02;
				buff[7] = 0;
				}
				SCpnt->result = (DID_OK << 16);
				SCpnt->scsi_done = done;
				done(SCpnt);
				break;
			case READ_6:
			case WRITE_6:
			case READ_10:
			case WRITE_10:
				error = ulsata2_ata_rw(SCpnt);
				break;
			case MODE_SENSE:		/* 0x1a */
				buff[0] = 0xf;		/* mode data length */
				buff[1] = 0;		/* default medium type */
				buff[2] = 0x10;		/* dpo/fua support on */
				buff[3] = 0;		/* no block descriptors */
				buff[4] = 0x1c;		/* caching page */
				buff[5] = 0xa;		/* page length */
				buff[6] = smartstatus ? 0x00 : 0x08 ;
				buff[7] = 0x04 ;
				SCpnt->result = (DID_OK << 16);
				SCpnt->scsi_done = done;
				done(SCpnt);
				break;
			case MODE_SELECT:		/* 0x15 */
			        /* prepare cam task */
			        queue = ulsata2_request_start();
			       	queue->buffer=buff;
			       	queue->buffer_size=bufflen;
			        queue->cam_task.bCh=(U8)(id / DrvPCh);
			        queue->cam_task.bID=(U8)(id % DrvPCh);
			        queue->cam_task.ATACmdFlag=ATA_DEVICE+ATA_STATUS_RETURN;
				if (buff[7] == 0x04)
					queue->cam_task.Cmd.ATACmd.bFeature = SMART_DISABLE;
				else if (buff[7] == 0x06)
					queue->cam_task.Cmd.ATACmd.bFeature = SMART_ENABLE;
				else
					queue->cam_task.Cmd.ATACmd.bFeature = smartstatus ? SMART_DISABLE : SMART_ENABLE ;
				queue->cam_task.Cmd.ATACmd.bCount = 0;
				queue->cam_task.Cmd.ATACmd.LBAl = 0x4f<<8 | 0xc2<<16;
				queue->cam_task.Cmd.ATACmd.LBAh = 0;
				queue->cam_task.Cmd.ATACmd.bCmd = WIN_SMART;
				queue->cam_task.SGCount = 0;
				queue->cam_task.DataTransferLength = (U32)queue->buffer_size;
				queue->cam_task.DataBuffer = (void *)queue->buffer;
				queue->cam_task.callback = ulsata2_cbk_ioctl;
				queue->buffer_size = 512;
	
				/* attach queue and backup cam_task */
				queue->cam_task.pTaskExtension = (void *)queue;
				memcpy(&queue->cam_task2, &queue->cam_task, 
					sizeof(ATA_TASK));
				
				if (CAM_SubmitATA(&queue->cam_task) != camACCEPTED) {
					printk(KERN_WARNING "%s:[warning] submit channel%2d busy\n",PDCNAME, queue->cam_task.bCh );
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
				spin_unlock_irqrestore(pada->ulsata2_lock, lock);
		#else	
				spin_unlock_irqrestore(&CAM_spinlock, lock);
		#endif			
				return(1);
				}
	
				if (queue->cam_task.Cmd.ATACmd.bFeature == SMART_ENABLE)
					smartstatus = 1 ;
				else
					smartstatus = 0 ;
	
				SCpnt->result = (DID_OK << 16);
				SCpnt->scsi_done = done;
				done(SCpnt);
				break;
			case REQUEST_SENSE:		/* 0x03 */
				SCpnt->result = (DID_OK << 16);
				SCpnt->scsi_done = done;
				done(SCpnt);
				break;
			default:
				if (debug & LOGATA) {
					printk(KERN_WARNING "%s:[warning] unknown scsi command: %#x\n",PDCNAME, cmd[0]);
				}
				SCpnt->result = (DID_BAD_TARGET << 16);
				done(SCpnt);
		}
	}

	if (error) {
		/* return and retry again */
		/* SCpnt->result = (BUSY<<1|COMMAND_COMPLETE<<8|DID_OK<<16); */
		//gibsond
		//SCpnt->result = DID_ERROR << 16;
		SCpnt->result = DID_ABORT << 16;
		done(SCpnt);
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	spin_unlock_irqrestore(pada->ulsata2_lock, lock);
#else	
	spin_unlock_irqrestore(&CAM_spinlock, lock);
#endif	
	return(0);
}

/*
 * scsi abort function
 */
int ulsata2_scsi_abort(Scsi_Cmnd *SCpnt)
{
       	//printk(KERN_INFO "%s:[info] scsi abort success\n",PDCNAME);
	//if (SCpnt->serial_number != SCpnt->serial_number_at_timeout) {
	//	return(SCSI_ABORT_NOT_RUNNING);
	//}
	//SCpnt->result = DID_ABORT << 16;
	//SCpnt->scsi_done(SCpnt);
	//return(SCSI_ABORT_SUCCESS);
// This is called when a scsi cmnd timeout at the os level
// Returning SCSI_ABORT_ERROR will cause the os to call our
// reset routine. This is how miniport handles timeouts.
   printk(KERN_INFO "%s:[info] scsi abort - running FAILED\n",PDCNAME);
   return(FAILED);
}
/*
 * scsi eh abort function
 */
int ulsata2_scsi_eh_abort(Scsi_Cmnd *SCpnt)
{
       	//printk(KERN_INFO "%s:[info] scsi abort success\n",PDCNAME);
	//if (SCpnt->serial_number != SCpnt->serial_number_at_timeout) {
	//	return(SCSI_ABORT_NOT_RUNNING);
	//}
	//SCpnt->result = DID_ABORT << 16;
	//SCpnt->scsi_done(SCpnt);
	//return(SCSI_ABORT_SUCCESS);
// This is called when a scsi cmnd timeout at the os level
// Returning SCSI_ABORT_ERROR will cause the os to call our
// reset routine. This is how miniport handles timeouts.
   printk(KERN_INFO "%s:[info] scsi eh abort - running FAILED\n",PDCNAME);
   return(FAILED);
}
/*
 * scsi reset function
 */
int ulsata2_scsi_reset(Scsi_Cmnd *SCpnt, unsigned int flags)
{
	unsigned long lock = 0;
	unsigned char chn;
	unsigned int id;
#ifndef KERNEL_VERSION_25x
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->host->hostdata;
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->target;
#else
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->device->host->hostdata;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->id;
#else
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->channel;
#endif	
#endif

	chn = (unsigned char)(id / DrvPCh);
       	printk(KERN_WARNING "%s:[warning] scsi reset channel%x ",PDCNAME,chn+1);
	SCpnt->result = DID_RESET << 16;
	
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		spin_lock_irqsave(pada->ulsata2_lock, lock);
		#else
		spin_lock_irqsave(&CAM_spinlock, lock);
		#endif	
	
	if (CAM_ResetATAChannel(chn, ulsata2_cbk_reset) == camSUCCESS) {
	
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		spin_unlock_irqrestore(pada->ulsata2_lock, lock);
		#else	
		spin_unlock_irqrestore(&CAM_spinlock, lock);
		#endif	
			
		printk("OK\n");
	       	return(SCSI_RESET_SUCCESS);
	}
	else {
		
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		spin_unlock_irqrestore(pada->ulsata2_lock, lock);
		#else	
		spin_unlock_irqrestore(&CAM_spinlock, lock);
		#endif		
				
		printk("Failed\n");
	       	return(SCSI_RESET_ERROR);
	}
}
/*
 * scsi eh device reset function
 */
int ulsata2_scsi_eh_device_reset(Scsi_Cmnd *SCpnt)
{
	unsigned char chn;
	unsigned int id;
	ulsata2_drive_t *pdrv;
	PDEV_CONFIG cam_dev;
	
#ifndef KERNEL_VERSION_25x
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->host->hostdata;
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->target;
#else
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->device->host->hostdata;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->id;
#else
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->channel;
#endif		
#endif

	chn = (unsigned char)(id / DrvPCh);
	pdrv = &ulsata2_drive[id];
	cam_dev = pdrv->cam_dev;
		
	//if(++errTimePCH[chn]>2)
	
	//printk(KERN_WARNING "%s:[warning] scsi channel reset disk%x ",PDCNAME,id+1);
	//if(ulsata2_scsi_reset(SCpnt,0) == SCSI_RESET_SUCCESS)
	//{	//errTimePCH[chn]=0;
	//	//return(SUCCESS);
	//	printk("OK\n");}else
	//	printk("Failed\n");
	printk(KERN_WARNING "%s:[warning] scsi eh reset disk%x ",PDCNAME,id+1);
	
	if (CAM_EnumATADevice(cam_dev) == camSUCCESS) {
		printk("OK\n");
	       	return(SUCCESS);
	}
	else {
		printk("Failed\n");
		return(FAILED);
	}
}
/*
 * scsi eh host reset function
 */
int ulsata2_scsi_eh_host_reset(Scsi_Cmnd *SCpnt)
{

	unsigned char chn;
	unsigned int id;
#ifndef KERNEL_VERSION_25x
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->host->hostdata;

	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->target;
#else
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)SCpnt->device->host->hostdata;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->id;
#else
	id = (unsigned int)pada->id*ChPAda*DrvPCh + SCpnt->device->channel;
#endif
#endif
	chn = (unsigned char)(id / DrvPCh);
       	printk(KERN_WARNING "%s:[warning] scsi eh reset channel%x ",PDCNAME,chn+1);
	if (CAM_ResetATAChannel(chn, ulsata2_cbk_reset) == camSUCCESS) {
		printk("OK\n");
	       	return(SUCCESS);
	}
	else {
		printk("Failed\n");
	       	return(FAILED);
	}
}
/*
 * returns unit geometry in cylinders/heads/sectors
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
int ulsata2_bios_param(struct scsi_device *disk, struct block_device *blkdev, sector_t sectnum, int Geometry[])
{
	unsigned long capacity = disk->capacity;
#else
int ulsata2_bios_param(struct scsi_device *sdev, struct block_device *bdev, sector_t cap, int Geometry[3])
{
	unsigned long capacity = (unsigned long)cap;
#endif

	Geometry[0]=64;  	
	Geometry[1]=32;  	
	Geometry[2]=(int)(capacity / (Geometry[0]*Geometry[1])); 
 
	/* capacity larger than 1GB(C/H/S = 1024/64/32) */
	if(capacity > 0x200000) {
		Geometry[0]=255;
		Geometry[1]=63; 
		Geometry[2]=(int)(capacity/(Geometry[0]*Geometry[1])); 
 
		/* capacity larger than 526GB(C/H/S = 65535/255/63) */
		if (Geometry[2] > 65535) {
			Geometry[2] = 65535;
			Geometry[1] = (int)(capacity / (255*65535)) ;
  		}  
	}

	return 0;
} 

/*
 * char dev operation
 */
static int ulsata2dev_open(struct inode *inodep, struct file *filep)
{
	
#ifndef KERNEL_VERSION_25x
	MOD_INC_USE_COUNT;
#endif
	return(0);
}
static int ulsata2dev_release(struct inode *inodep, struct file *filep)
{
#ifndef KERNEL_VERSION_25x
	MOD_DEC_USE_COUNT;
#endif
	return(0);
}
static int ulsata2dev_ioctl(struct inode *inodep, struct file *filep, 
		unsigned int req, unsigned long arg)
{
	int retval = -1;
	unsigned char *args;

	if ( !arg ) {
		return retval;
	}

	args = (unsigned char *)arg;
	switch (req) {
		/* get total adapter count */
		case STIO_GET_ADAPTERCOUNT: {
		tag_adaptercount_t *buf = (tag_adaptercount_t *)args;

		buf->adaptercount = (unsigned char)ulsata2_num_adapters;

		retval = 0;
		break;
		}

		/* get adapter type */
		case STIO_GET_ADAPTERTYPE: {
		tag_adaptertype_t *buf = (tag_adaptertype_t *)args;
		int ada = (int)buf->adapter;

		if (((ulsata2_adapter[ada].present)&ulsata2_present)) {
			switch (ulsata2_adapter[ada].deviceid) {
				case DID518:
					buf->adaptertype = SATA518;
					break;
				case DID575:
					buf->adaptertype = SATA575;
					break;
				case DID579:
					buf->adaptertype = SATA579;
					break;
				case DID718:
					buf->adaptertype = SATA718;
					break;
				case DID775:
					buf->adaptertype = SATA775;
					break;
				case DID779:
					buf->adaptertype = SATA779;
					break;
				default:
					buf->adaptertype = SATA150;
					break;
			}
			retval = 0;
		}

		break;
		}

		/* get adapter info */
		case STIO_GET_ADAPTERINFO: {
		tag_adapterinfo_t *buf = (tag_adapterinfo_t *)args;
		int ada = (int)buf->adapter;

		if (((ulsata2_adapter[ada].present)&ulsata2_present)) {
			buf->deviceid = ulsata2_adapter[ada].deviceid;
			retval = 0;
		}

		break;
		}
		/* find flash mem type */
		case STIO_FIND_FLASHMEM: {
		tag_flash_t *buf = (tag_flash_t *)args;
		int ret;

		ret = (int)CAM_Find_FlashMemory(buf->adapter, buf->buffer);

		if (ret ==camFAIL) {
			return -1;
		}

		retval = 0;
		break;
		}

		/* erase flash memory */
		case STIO_ERASE_FLASHMEM: {
		int ret;
		tag_flash_t *buf = (tag_flash_t *) args;

		ret = (int)CAM_Erase_FlashMemory(buf->adapter);
		if (ret == camSUCCESS) {
			retval = 0;
		}

		break;
		}

		/* read flash memory */
		case STIO_READ_FLASHMEM: {
		int ret;
		tag_flash_t *buf = (tag_flash_t *) args;

		ret = (int)CAM_Read_FlashMemory(buf->adapter, buf->offset, buf->buffer_size, buf->buffer);
		if (ret == camSUCCESS) {
			retval = 0;
			if (debug & LOGATA) {
				printk(KERN_DEBUG "%s:[D] read %ldK offset %ld success ", PDCNAME, buf->buffer_size, buf->offset);
				hexdump(buf->buffer, 10);
			}
		}

		break;
		}

		/* write flash memory */
		case STIO_WRITE_FLASHMEM: {
		int ret;
		tag_flash_t *buf = (tag_flash_t *) args;

		ret = (int)CAM_Write_FlashMemory(buf->adapter, buf->offset, buf->buffer_size, buf->buffer);
		if (ret == camSUCCESS) {
			retval = 0;
			if (debug & LOGATA) {
				printk(KERN_DEBUG "%s:[D] write %ldK offset %ld success ", PDCNAME, buf->buffer_size, buf->offset);
				hexdump(buf->buffer, 10);
			}
		}

		break;
		}
		default:
			break;
	}
	return retval;
}
/*
 * Implement CAM Supportive Routines
 */
//void camOutPortByte(U32 port, U8 data)
void camOutPortByte(unsigned long port, U8 data)
{
	outb(data, port);
}
//U8 camInPortByte(U32 port)
U8 camInPortByte(unsigned long port)
{
	return inb(port);
}
//void camOutPortWord(U32 port, U16 data)
void camOutPortWord(unsigned long port, U16 data)
{
	outw(data, port);
}
//U16 camInPortWord(U32 port)
U16 camInPortWord(unsigned long port)
{
	return inw(port);
}

//void camOutPortDword(U32 port, U32 data)
void camOutPortDword(unsigned long port, U32 data)
{
	outl(data, port);
}
//U32 camInPortDword(U32 port)
U32 camInPortDword(unsigned long port)
{
	return inl(port);
}
/*
 * MMIO function for CAM
 */
//void camWriteRegByte(U32 addr, U8 data)
void camWriteRegByte(unsigned long addr, U8 data)
{
	writeb(data, addr);
}
//U8 camReadRegByte(U32 addr)
U8 camReadRegByte(unsigned long addr)
{
	return readb(addr);
}
//void camWriteRegWord(U32 addr, U16 data)
void camWriteRegWord(unsigned long addr, U16 data)
{
	writew(data, addr);
}
//U16 camReadRegWord(U32 addr)
U16 camReadRegWord(unsigned long addr)
{
	return readw(addr);
}
//void camWriteRegDword(U32 addr, U32 data)
void camWriteRegDword(unsigned long addr, U32 data)
{
	writel(data,addr);
}
//U32 camReadRegDword(U32 addr)
U32 camReadRegDword(unsigned long addr)
{
	return readl(addr);
}

/*
 * Function:	camCheckDriveStatus()
 *
 * Purpose:	Report to driver if the drive is plugged in or unplugged
 *
 * Parameters:	bCh	- channel # (0 - N)
 * 		bID	- 0 is master, 1 is slave
 * 		Status	- NEW_DRIVE: a new drive is plugged in
 * 			  NO_DRIVE:  the drive is unplugged
 *
 * Return:	nothing
 *
 * Notes:	use tasklet to scan device.
 * 		FIXME: there seems no function to notify scsi when disk is 
 * 		       unplugged in kernel 2.4.x
 */
DECLARE_TASKLET(tasklet, ulsata2_init_device, 0);

void camCheckDriveStatus(U8 bCh, U8 bID, U8 Status)
{
	int drv, ret;
	ulsata2_drive_t *pdrv;
	DEV_CONFIG *pcam_dev; 

	drv = (int)bCh*DrvPCh+(int)bID;
	pdrv = &ulsata2_drive[drv];
	pcam_dev = pdrv->cam_dev;

	/* SATA disk unplug */
	if (Status == NO_DRIVE) {
		ENCLOSURE_TYPE	type;
		pcam_dev->DevFlag = DEV_NOTFOUND;
#ifdef	SSBOX
		if ( ( ret = CAM_EnumEnclosure(bCh, &type) ) != camNOBOX ) {
			SwapBox_CtrlLED(bCh, LED_Red, LEDNOCHANGE, StatusLED );
			SwapBox_CtrlLED(bCh, LEDOFF, LEDNOCHANGE, ActiveLED );
		}
#endif
		printk(KERN_INFO "%s:[info] Drive%2d Offline: %s \n",PDCNAME, drv+1, pcam_dev->Model);
	}
	/* new SATA disk plugin */
	else if (Status == NEW_DRIVE) {
		printk(KERN_INFO "%s:[info] Drive%2d PlugIn: \n",PDCNAME, drv+1);
		//gibsond
		//ulsata2_init_device((unsigned long)drv);

		tasklet.data = (unsigned long)drv; //Beware this!!!!
		tasklet_schedule(&tasklet);
	}
	return;
}

/*
 * Function:	camStallExecution()
 *
 * Purpose:	provid CAM a OS-related interface to stalls under the OS.
 *
 * Parameters:	Delay	- specifies the delay interval in 10(ms)
 *
 * Return:	nothing
 *
 * Notes:	n/a
 */
void camStallExecution(U32 Delay)
{
	unsigned int i=0,j=0;	
	//mdelay((unsigned long)Delay*10);
	for(i=0;i<1000;i++)
	if(Delay<=10)
		udelay((unsigned long)Delay*10);
	else	
		for(j=0;j<(Delay%10);j++)
		udelay(100);
	return;
}

/*
 * Function:	camUStallExecution()
 *
 * Purpose:	provid CAM a OS-related interface to stalls under the OS.
 *
 * Parameters:	Delay	- specifies the delay interval in 10(us)
 *
 * Return:	nothing
 *
 * Notes:	n/a
 */
void camUStallExecution(U32 Delay)
{
	unsigned int i=0;
	if(Delay<=10)
		udelay((unsigned long)Delay*10);
	else	
		for(i=0;i<(Delay%10);i++)
		udelay(100);
	
	return;
}

/*
 * Function:	ulsata2_pre_cam_timer()
 *
 * Purpose:	when ATA timeout, call pre timer before TimerService to do some
 * 		task
 *
 * Parameters:	pointer to ulsata2_timer_t
 *
 * Return:	nothing
 *
 * Notes:	when ATA timeout, call our pre timer then call cam TimerService.
 * 		when timeout, cam won't del timer, thus status will always be 
 * 		BUSY.
 */
void ulsata2_pre_cam_timer(unsigned long ptr)
{
	ulsata2_timer_t *t = (ulsata2_timer_t *)ptr;
	unsigned long	lock;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	ulsata2_adapter_t  *pada = &ulsata2_adapter[0];
	spin_lock_irqsave(pada->ulsata2_lock, lock);
#else	
	spin_lock_irqsave(&CAM_spinlock, lock);
#endif	
	//panic(KERN_ERR "ulsata2 timeout panic!!\n");

	t->status = camFREE;
	del_timer(&t->timer);
	t->callback();
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	spin_unlock_irqrestore(pada->ulsata2_lock, lock);
#else	
	spin_unlock_irqrestore(&CAM_spinlock, lock);
#endif	
	return;
}

/*
 * Function:	camTimer()
 *
 * Purpose:	export OS timer for CAM
 *
 * Parameters:	AdapterNumber	- Adapter ID
 *              Delay		- the timeout (ticks=1/18 second)
 * 		TimerService	- timer expire function
 *
 * Return:	timer index
 *
 * Notes:	don't return 0 as index
 */
#ifdef	_64BIT_
unsigned long camTimer(U8 AdapterNumber, U32 Delay, void *TimerService)
#else
U32 camTimer(U8 AdapterNumber, U32 Delay, void *TimerService)
#endif
{
	unsigned char i;
	unsigned long flags;
	//ulsata2_adapter_t	*pada = &ulsata2_adapter[ut_adapters];
	
	//spin_lock_irqsave(pada->timer_lock, flags);
	
	if (Delay) {
		for (i = 0; i < MAXTIMER; i++) {
			/* add timer */
			if ( ulsata2_timer[i].status == camFREE ) {
				memset(&ulsata2_timer[i], 0, 
						sizeof(ulsata2_timer_t));
				init_timer(&ulsata2_timer[i].timer);
				ulsata2_timer[i].status = camBUSY;
				ulsata2_timer[i].timer.expires = (unsigned long)jiffies + ((unsigned long)Delay*HZ/18);
				ulsata2_timer[i].timer.function = ulsata2_pre_cam_timer;
				ulsata2_timer[i].timer.data = (unsigned long)&ulsata2_timer[i];
				ulsata2_timer[i].id = i;
				ulsata2_timer[i].callback = TimerService;
				add_timer(&ulsata2_timer[i].timer);
				/* don't return 0 as index */
				//spin_unlock_irqrestore(pada->timer_lock, flags);
				return(i+1);
			}
		}
	}
	else {	/* remove timer */
		int num;
		num = (int)TimerService - 1;
		if (ulsata2_timer[num].status == camBUSY)
			del_timer(&ulsata2_timer[num].timer);
		ulsata2_timer[num].status = camFREE;
		//spin_unlock_irqrestore(pada->timer_lock, flags);
		return(0);
	}
	
	//spin_unlock_irqrestore(pada->timer_lock, flags);
	return(0);
}
/*
 * SwapBox Support
 */
void
SwapBox_CtrlLED( U8 channel, U8 Status, U8 Flash, U8 Type )
{
	CAM_Ctrl_LED Ctrl_LED;
	Ctrl_LED.LED.StatusLed=Status;
	Ctrl_LED.Flash=Flash;
	Ctrl_LED.LED_Type=Type;
	CAM_CtrlLED( channel,&Ctrl_LED );
}

void
SwapBox_CtrlLEDEmptyDevices( U8 AdapterId, U8 Status, U8 Flash, U8 Type )
{
	CAM_Ctrl_LED Ctrl_LED;
	Ctrl_LED.LED.StatusLed=Status;
	Ctrl_LED.Flash=Flash;
	Ctrl_LED.LED_Type=Type;
	CAM_InitWithoutDiskSATAEnclosures( AdapterId,&Ctrl_LED );
}

/*
 * Module Stuff
 */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10)
MODULE_AUTHOR("Promise Technology Inc.");
#ifdef __SMP__
MODULE_DESCRIPTION("Promise SATAII150 Series Linux Driver (SMP)");
#else
MODULE_DESCRIPTION("Promise SATAII150 Series Linux Driver");
#endif
MODULE_LICENSE("GPL");
#endif
//MODULE_PARM(atapi, "i");
//MODULE_PARM(debug, "i");
//MODULE_PARM_DESC(atapi, "ATAPI enable bit (1/0, default 0)");
//MODULE_PARM_DESC(debug, "The debug level (0-2)");

////////////////////////////////////////////////////////////////////////////////
#define ulsata2{							\
	.module				= THIS_MODULE,		\
	.proc_dir 			= NULL,			\
	.proc_info			= ulsata2_proc_info,			\
	.name				= "ulsata2",		\
	.detect				= ulsata2_detect,		\
	.release			= ulsata2_release,	\
	.ioctl				= ulsata2_ioctl,			\
	.info				= NULL,			\
	.queuecommand			= ulsata2_queuecommand,	\
	.eh_abort_handler 		= ulsata2_scsi_eh_abort,	\
	.eh_device_reset_handler 	= ulsata2_scsi_eh_device_reset,	\
	.bios_param			= ulsata2_bios_param,	\
	.can_queue			= QLENGTH,		\
	.this_id			= -1,			\
	.sg_tablesize			= MAXSGTABLE,	\
	.cmd_per_lun			= 1,			\
	.present			= 0,			\
	.unchecked_isa_dma		= 0,			\
	.use_clustering			= 0,			\
        .emulated			= 0, 			\
        .proc_name			= "ulsata2", 		\
} 

#ifdef MODULE
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0) 
Scsi_Host_Template driver_template = ulsata2;
/* Only export this gobal variable in system symbol table */
/* we shall add 3 items in MAkefile: -DMODVERSIONS, -DEXPORT_SYMTAB, -include linux/modversions.h  */
EXPORT_SYMBOL(driver_template);
/*
 * unregister scsi host when unregister scsi module
 */
int ulsata2_release(struct Scsi_Host *shpnt)
{
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)shpnt->hostdata;
	static int release_count;

	/* close cam */
       	if(((pada->present)&ulsata2_present)) {
	       	ADAPTER_CONFIG_INFO *pcam_ada = pada->cam_ada;
#ifdef SSBOX
		CAM_CloseSATAEnclosure(pada->id);
		if (swapbox_exist)
			del_timer(&swapbox_timer);
#endif
	       	if (CAM_Close(pcam_ada) != camSUCCESS )
		       	printk(KERN_ERR "%s:[error] adapter%x cam close fail\n",PDCNAME,pada->id);
       	}

	/* free memory, irq, ioport */
	ulsata2_cleanup(pada->id);

	if (ulsata2_num_adapters == ++release_count) {
		/* free queue, timer */
		ulsata2_cleanup2();

		/* unregister chr device */
		unregister_chrdev(ulsata2_major, "ulsata2");
	}

	/* unregister scsi host */
	scsi_unregister(shpnt);
	
	/* fake event to notify shutdown */
	ulsata2_halt(0, SYS_DOWN, 0);
	return(0);
}
#include <drivers/scsi/scsi_module.c>
#else	/* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) */

int ulsata2_notifier_func(struct notifier_block *pnb,unsigned long code, void *unused)
{
	printk("notifier_func\n");
	return(NOTIFY_DONE);
}

/*
 * unregister scsi host when unregister scsi module
 */
int ulsata2_release(struct Scsi_Host *shpnt)
{
	ulsata2_adapter_t *pada =(ulsata2_adapter_t *)shpnt->hostdata;
	static int release_count;

	/* close cam */
       	if(((pada->present)&ulsata2_present)) {
	       	ADAPTER_CONFIG_INFO *pcam_ada = pada->cam_ada;
#ifdef SSBOX
		CAM_CloseSATAEnclosure(pada->id);
		if (swapbox_exist)
			del_timer(&swapbox_timer);
#endif
	       	if (CAM_Close(pcam_ada) != camSUCCESS )
		       	printk(KERN_ERR "%s:[error] adapter%x cam close fail\n",PDCNAME,pada->id);
       	}

	/* free memory, irq, ioport */
	ulsata2_cleanup(pada->id);

	if (ulsata2_num_adapters == ++release_count) {
		/* free queue, timer */
		ulsata2_cleanup2();

		/* unregister chr device */
		unregister_chrdev(ulsata2_major, "ulsata2");
	}

	/* unregister scsi host */
	scsi_unregister(shpnt);
	
	/* fake event to notify shutdown */
	ulsata2_halt(0, SYS_DOWN, 0);
	return(0);
}
static Scsi_Host_Template driver_template = ulsata2;

static void ulsata2_iounmap(ulsata2_adapter_t *pada)
{
int i;
//for(i = 0; i < BASE_ADDRESS_MAX_NUM; i++)
for(i = 0; i < 6; i++)
	//if(p->bitmap_ioremap & (1<<i))
if(pci_resource_flags(pada->pci_dev, i) & IORESOURCE_MEM) 
			iounmap((void *)pada->base[i]);
//p->bitmap_ioremap = 0;
}                                                  
static int release_single_device(ulsata2_adapter_t *pada)
{
	static int release_count=0;
	int i=0;
	//free_irq(pada->irq, pada);
	//ulsata2_iounmap(pada);
	/* close cam */
       	if(((pada->present)&ulsata2_present)) {
	       	ADAPTER_CONFIG_INFO *pcam_ada = pada->cam_ada;
#ifdef SSBOX
		CAM_CloseSATAEnclosure(pada->id);
		if (swapbox_exist)
			del_timer(&swapbox_timer);
#endif
	       	if (CAM_Close(pcam_ada) != camSUCCESS )
		       	printk(KERN_ERR "%s:[error] adapter%x cam close fail\n",PDCNAME,pada->id);
       	}
	
	printk("ulsata2 cards=%d\n",ulsata2_num_adapters);
	if(ulsata2_num_adapters == 1)	
		for(i=0;i<2;i++)
			ulsata2_cleanup(i);
	else
			ulsata2_cleanup(pada->id);
	if (ulsata2_num_adapters == ++release_count)	
	ulsata2_cleanup2();
	pci_release_regions(pada->pci_dev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)	
	scsi_host_put(pada->host);
#endif
	pci_disable_device(pada->pci_dev);	
	pci_set_drvdata(pada->pci_dev, NULL);
	
	return 1;
}
/**
* ulsata2_probe - Adapter hot plug add entry point
*
* Return value:
* 	0 on success / non-zero on failure
**/
static int __devinit ulsata2_probe(struct pci_dev *pdev,
			       const struct pci_device_id *dev_id)
{
	unsigned char ua;
	//unsigned char adapters=0;//, adapcnt;
	//unsigned char class;
	struct Scsi_Host *shpnt;
	u32 rc=PCIBIOS_SUCCESSFUL;	
	unsigned int re_size;
	
	printk(" Enter ulsata2_probe \n");	
	//re_size = sizeof(ulsata2_adapter);	
	/*Step1: enable and initialize pci device*/
	if ((rc = pci_enable_device(pdev))) {
		dev_err(&pdev->dev, "Cannot enable adapter\n");
		return rc;
	}

	rc = pci_request_regions(pdev, "ulsata2");
	if (rc)
		goto err_out;
	
	/*enabled dma mode*/
	rc=pci_set_dma_mask(pdev, 0xffffffffffffffff);
	if (rc)
		goto err_out_regions;
	pci_set_master(pdev);

	dev_info(&pdev->dev, "Found OII Controller with IRQ: %d\n", pdev->irq);

	/*Step2: alloc and initialize scsi host */

	shpnt = scsi_host_alloc(&driver_template, sizeof(ulsata2_adapter_t));		

	if (!shpnt) {
		dev_err(&pdev->dev, "call to scsi_host_alloc failed!\n");
		rc = -ENOMEM;
		goto err_out_regions;
	}

	shpnt->max_channel=0;
	shpnt->max_id=DrvPCh*ChPAda;
	shpnt->max_lun=0;
	shpnt->unique_id = pdev->irq;
	shpnt->max_cmd_len= 15;//16;	
	shpnt->unchecked_isa_dma=0;
	shpnt->irq = pdev->irq;
	//shpnt->host_lock=&CAM_spinlock;
	ulsata2_adapter_t	*pada = &ulsata2_adapter[ut_adapters];
	pada->present &= ~ulsata2_present;
	/* initialize global data structure */
	if(ulsata2_init_gds()) {
	       	printk(KERN_ERR "%s:[error] initialize global data structure failed\n",PDCNAME);
		rc = -ENOMEM;
		goto err_out_regions;
	}
	
	
	if (shpnt->hostdata) {
			pada->host = shpnt;
			pada->present |= ulsata2_present;
			pada->pci_dev = pdev;
			pada->deviceid = pdev->device;
			pada->irq = pdev->irq;
			pada = (ulsata2_adapter_t *) shpnt->hostdata;
	 	      	memcpy(pada, &ulsata2_adapter[ut_adapters], sizeof(ulsata2_adapter_t));
	 	      	
		} else {
		       	printk(KERN_ERR "%s:[error] adapter%d: register scsi host failed\n",PDCNAME,ua);
		       	goto err_out_regions;
	       	}  		
	
	
	pci_set_drvdata(pdev, pada);

		
	/* Step5: register resources and interrupt handler */
	if(init_pci_resource(pada, pdev) ){
		rc = -ENOMEM;
		goto err_out_uncach;
	}	

	rc = scsi_add_host(shpnt, &pdev->dev);
	if (rc){
		//printk("%s:[error] adapter%d request irq0x%x failed\n",PDCNAME, pada, pdev->irq);
		goto err_out_iounmap;
	}
		
/* initialize cam */
	if (ulsata2_init_cam(ut_adapters)) {
		       	printk(KERN_ERR "%s:[error] adapter%d: init cam failed\n",PDCNAME,ut_adapters);
		       	rc = -ENOMEM;
			goto err_out_freeirq;
	       	}
	
	ulsata2_num_adapters=++ut_adapters;
	printk("ulsata2_lock=%p\n",pada->ulsata2_lock);
	if(pada->ulsata2_lock<0xffff){
		goto err_out_freeirq;
	}
	scsi_scan_host(shpnt);

#ifdef SSBOX
	/* SwapBox Monitor Timer */
	if ( swapbox_exist ) {
		init_timer(&swapbox_timer);
		swapbox_timer.expires = jiffies + 30 * HZ ;
		swapbox_timer.data = 0 ;
		swapbox_timer.function = update_swapbox ;
		add_timer(&swapbox_timer);
	}
#endif
	
	printk(KERN_INFO ":exit ulsata2_probe\n");
	
	return 0;

err_out_freeirq:
	free_irq(pdev->irq, pada);
err_out_iounmap:
	release_region(pada->base[0],pada->range[0]);
	release_region(pada->base[2],pada->range[2]);
err_out_uncach:
err_out_vmalloc:
	for(ua=0;ua<2;ua++)
		ulsata2_cleanup(ua);
	ulsata2_cleanup2();
err_out_host:
	scsi_host_put(shpnt);
err_out_regions:
	pci_release_regions(pdev);
err_out:
	pci_disable_device(pdev);

	return rc;

}


/**
 * ulsata2_remove - adapter hot plug remove entry point
 * @pdev:	pci device struct
 *
 * Adapter hot plug remove entry point.
 *
 * Return value:
 * 	none
 **/
 static void __devexit ulsata2_remove(struct pci_dev *pdev)
{
	ulsata2_adapter_t *pada = (ulsata2_adapter_t *)pci_get_drvdata(pdev);
	printk(KERN_INFO ":enter ulsata2_remove\n");

	scsi_remove_host(pada->host);	
	
	release_single_device(pada);

	
}
/**
 * ulsata2_shutdown - Shutdown handler.
 * @dev:	device struct
 *
 * This function is invoked upon system shutdown/reboot. It will issue
 * an adapter shutdown to the adapter to flush the write cache.
 *
 * Return value:
 * 	none
 **/
static void ulsata2_shutdown(struct device *dev)
{
	ulsata2_adapter_t *pada = (ulsata2_adapter_t *)pci_get_drvdata(to_pci_dev(dev));

	release_single_device(pada);
}
//static Scsi_Host_Template driver_template = ulsata2;
///*
//*	Supported Hardware
//*/
static struct pci_device_id ulsata2_pci_table[] __devinitdata = {
	{ PCI_VENDOR_ID_PROMISE, DID575, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0UL }, 
	{ PCI_VENDOR_ID_PROMISE, DID518, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0UL },
	{ PCI_VENDOR_ID_PROMISE, DID579, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0UL }, 
	{ PCI_VENDOR_ID_PROMISE, DID775, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0UL },
	{ PCI_VENDOR_ID_PROMISE, DID718, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0UL }, 
	{ PCI_VENDOR_ID_PROMISE, DID779, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0UL }, 
	{}
};


static struct pci_driver ulsata2_driver = {
	.name = "ulsata2",
	.id_table = ulsata2_pci_table,
	.probe = ulsata2_probe,
	.remove = ulsata2_remove,
	.driver = {
		.shutdown = ulsata2_shutdown,
	},	
};

/**
 * ulsata2_init - Module entry point
 *
 * Return value:
 * 	0 on success / non-zero on failure
 **/
static int __init ulsata2_init(void)
{
	 

	printk("PROMISE %s Series Linux Driver v%s\n", PRODUCT, VERSION);

	if ((ulsata2_major = register_chrdev(0, "ulsata2", 
						&ulsata2_fops)) < 0 ) {
			printk(KERN_ERR "%s:[error] register character device failed\n",PDCNAME);
		}
	
//#ifdef _64BPLATFORM
//        register_ioctl32_conversion(IOCTL_FASTTRAK_GETVERSION, NULL);
//        register_ioctl32_conversion(IOCTL_FASTTRAK_PASSTHRU, NULL);
//#endif /* _64BPLATFORM */
        
        ///* initialize global data structure */
	//if(ulsata2_init_gds()) {
	//       	printk(KERN_ERR "%s:[error] initialize global data structure failed\n",PDCNAME);
	//	return -1;
	//}
        
        
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
	if (!pci_register_driver(&ulsata2_driver) ) {
		printk("pci_unregister_driver\n");
               	pci_unregister_driver(&ulsata2_driver);
        	return -ENODEV;
    	}
                     
        return 0;
#else        
	return pci_register_driver(&ulsata2_driver);
#endif /* KERNEL_VERSION_CODE < KERNEL_VERSION(2.6.10) */
}
/**                                                             
 * ulsata2_exit - Module unload                                  
 *                                                              
 * Module unload entry point.                                   
 *                                                              
 * Return value:                                                
 * 	none                                                    
 **/                                                            
static void __exit ulsata2_exit(void)                            
{                                                               
	if(ulsata2_major>0)                                      
		unregister_chrdev(ulsata2_major,"ulsata2");
	                                                        
	pci_unregister_driver(&ulsata2_driver);                  
	printk(KERN_ERR "exit ulsata2_exit\n");                  
}     

MODULE_DEVICE_TABLE(pci,ulsata2_pci_table);			/*needed for pci device*/


module_init(ulsata2_init);
module_exit(ulsata2_exit);
#endif
#endif
void	camDataCacheForceWrite( void )
{
	flush_cache_all();
}
